#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import json
import ctypes
import numpy as np

from enum import Enum
from tempfile import TemporaryDirectory
from neural_compressor.utils.utility import LazyImport
from neural_compressor.model.model import MXNetModel as NCModel

mx = LazyImport("mxnet")


QUANTIZE_OP_NAME = 'quantize_output'
QUANTIZE_OP_NAMES = ['_contrib_quantize_v2']
QUANTIZE_DEFAULT_ALGORITHM = 'minmax'
QUANTIZE_NODE_POSTFIX = '_quantize'
QUANTIZED_NODE_PREFIX = 'quantized_'
QUANTIZATION_DTYPES = [np.int8, np.uint8]
NULL_OP_NAMES = ['null']


class OpType(Enum):
    NORMAL = 0
    QUANTIZE = 1
    QUANTIZED = 2


def isiterable(obj) -> bool:
    """Checks whether object is iterable.

    Args:
        obj : object to check.

    Returns:
        boolean: True if object is iterable, else False.
    """
    try:
        iter(obj)
        return True
    except TypeError:
        return False


def ensure_list(x):
    """Ensures that object is a list.
    """
    return x if isinstance(x, (tuple, list)) else [x]


def check_mx_version(version):
    """Checks MXNet version.

    Args:
        version (str): version to check.

    Returns:
        boolean: True if mx.__version__ >= version, else False.
    """
    d1 = re.split(r'\.', mx.__version__)
    d2 = re.split(r'\.', version)
    d1 = [int(d1[i]) for i in range(len(d1))]
    d2 = [int(d2[i]) for i in range(len(d2))]
    return d1 >= d2


def combine_capabilities(current, new):
    if isinstance(current, list):
        assert isinstance(new, list)
        return list(set(current) | set(new))
    assert isinstance(current, dict) and isinstance(new, dict)
    current = current.copy()
    new = new.copy()
    for k, v in current.items():
        current[k] = combine_capabilities(v, new.get(k, type(v)()))
        new.pop(k, None)
    current.update(new)
    return current


def make_nc_model(target, sym_model, ctx, input_desc):
    """Converts a symbolic model to an Neural Compressor model.

        Args:
            target (object): target model type to return.
            sym_model (tuple): symbol model (symnet, args, auxs).
            input_desc (list): model input data description.

        Returns:
            NCModel: converted neural_compressor model
    """
    assert isinstance(sym_model, tuple) and isinstance(sym_model[0], mx.symbol.Symbol)

    if isinstance(target.model, mx.gluon.HybridBlock):
        return NCModel(make_symbol_block(sym_model, ctx, input_desc))
    return NCModel(sym_model)


def fuse(sym_model, ctx):
    """Fuse the supplied model.

        Args:
            sym_model (tuple): symbol model (symnet, args, auxs).

        Returns:
            tuple: fused symbol model (symnet, args, auxs).
    """
    assert isinstance(sym_model, tuple) and isinstance(sym_model[0], mx.symbol.Symbol)

    symnet, args, auxs = sym_model
    backend = get_backend_name(ctx)
    if backend is not None:
        if check_mx_version('2.0.0'):
            symnet = symnet.optimize_for(backend)
        else:
            symnet = symnet.get_backend_symbol(backend)
    return (symnet, args, auxs)


def get_backend_name(ctx):
    if 'cpu' in ctx.device_type:
        if check_mx_version('2.0.0'):
            return 'ONEDNN_QUANTIZE'
        else:
            return 'MKLDNN_QUANTIZE'
    return None


def prepare_model_data(nc_model, ctx, data_x):
    """Prepares sym_model and dataloader needed for quantization, calibration or running.

    Args:
        nc_model (object): model to prepare.
        data_x (object): data iterator/loader to prepare.

    Returns:
        tuple: symbol model (symnet, args, auxs) and DataLoaderWrap.
    """
    dataloader = prepare_dataloader(nc_model, ctx, data_x)
    sym_model = prepare_model(nc_model, ctx, dataloader.input_desc)
    return sym_model, dataloader


def prepare_model(nc_model, ctx, input_desc):
    assert isinstance(nc_model, NCModel)

    model_x = nc_model.model
    if isinstance(model_x, mx.gluon.HybridBlock):
        if not model_x._active:
            model_x.hybridize(static_alloc=False, static_shape=False)
        model_x(*create_data_example(ctx, input_desc))
        with TemporaryDirectory() as tmpdirname:
            prefix = os.path.join(tmpdirname, 'tmp')
            model_x.export(prefix, epoch=0, remove_amp_cast=False)
            sym_model = mx.model.load_checkpoint(prefix, 0)
    elif isinstance(model_x, tuple) and isinstance(model_x[0], mx.symbol.Symbol):
        sym_model = model_x
    else:
        raise TypeError('Wrong model type')
    if not is_model_quantized(sym_model):
        sym_model = fuse(sym_model, ctx)
    return _match_array_semantics(sym_model)


def create_data_example(ctx, input_desc):
    if mx.is_np_array():
        return [mx.np.zeros(d.shape, ctx=ctx, dtype=d.dtype) for d in input_desc]
    else:
        return [mx.nd.zeros(d.shape, ctx=ctx, dtype=d.dtype) for d in input_desc]


def _match_array_semantics(sym_model):
    if check_mx_version('2.0.0') and mx.util.is_np_array():
        symnet, args, auxs = sym_model
        symnet = symnet.as_np_ndarray()
        for k, v in args.items():
            args[k] = v.as_np_ndarray()
        for k, v in auxs.items():
            auxs[k] = v.as_np_ndarray()
        sym_model = (symnet, args, auxs)
    return sym_model


def prepare_dataloader(nc_model, ctx, data_x):
    assert isinstance(nc_model, NCModel)

    if isinstance(data_x, DataLoaderWrap):
        return data_x

    dataloader = data_x
    if isinstance(dataloader, mx.io.DataIter):
        dataloader = DataIterLoader(dataloader)
    assert isiterable(dataloader), 'Dataloader must be iterable (mx.gluon.data.DataLoader-like)'

    model_x = nc_model.model
    if isinstance(model_x, mx.gluon.HybridBlock):
        data = ensure_list(next(iter(dataloader)))  # data example
        data = [ndarray_to_device(d, ctx) for d in data]
        if not model_x._active:
            model_x.hybridize(static_alloc=False, static_shape=False)
        while True:
            try:
                model_x(*data)
            except (ValueError, TypeError):
                del data[-1]  # remove label
            else:
                break
        inputs, _ = model_x._cached_graph
        input_desc = [mx.io.DataDesc(name=i.name, shape=d.shape, dtype=d.dtype)
                      for i, d in zip(inputs, data)]
    elif isinstance(model_x, tuple) and isinstance(model_x[0], mx.symbol.Symbol):
        assert hasattr(data_x, 'provide_data'), \
            'Dataloader must provide data information (mx.io.DataDesc for each input)'
        input_desc = data_x.provide_data
    else:
        raise TypeError('Wrong model type')
    return DataLoaderWrap(dataloader, input_desc)


def ndarray_to_device(ndarray, device):
    try:
        return ndarray.to_device(device)  # 2.x version
    except AttributeError:
        return ndarray.as_in_context(device)  # 1.x version


def is_model_quantized(sym_model):
    """Checks whether the model is quantized.

        Args:
            sym_model (tuple): symbol model (symnet, args, auxs).
        Returns:
            boolean: True if model is quantized, else False.
    """
    assert isinstance(sym_model, tuple) and isinstance(sym_model[0], mx.symbol.Symbol)

    for sym in sym_model[0].get_internals():
        _, op_type = _dequantize_sym_name(sym.name)
        if op_type != OpType.NORMAL:
            return True
    return False


def _dequantize_sym_name(sym_name, check_list=None):
    name = sym_name
    op_type = OpType.NORMAL
    if sym_name.endswith(QUANTIZE_NODE_POSTFIX):
        op_type = OpType.QUANTIZE
        name = sym_name[:-len(QUANTIZE_NODE_POSTFIX)]
    elif sym_name.startswith(QUANTIZED_NODE_PREFIX):
        op_type = OpType.QUANTIZED
        name = sym_name[len(QUANTIZED_NODE_PREFIX):]
        assert check_list is None or name in check_list, \
            'name of the quantized symbol must be in the following format: ' \
            '"{}_<fp32_sym_name>". Symbol: {}'.format(QUANTIZED_NODE_PREFIX, name)
    return (name, op_type)


def query_quantizable_nodes(sym_model, ctx, dataloader):
    """Query quantizable nodes of the given model.

    Args:
        sym_model (tuple): symbol model (symnet, args, auxs) to query.

    Returns:
        list: quantizable nodes of the given model.
        dict: tensor to node mapping.
    """
    assert isinstance(sym_model, tuple) and isinstance(sym_model[0], mx.symbol.Symbol)
    assert isinstance(dataloader, DataLoaderWrap)

    symnet = sym_model[0]
    nodes_ops = {n['name']: n['op'].lower() for n in json.loads(symnet.tojson())['nodes']}

    qmodel, calib_tensors = quantize_sym_model(sym_model, ctx, {'quantized_dtype': 'auto',
                                                                'quantize_mode': 'smart'})
    qsymnet = qmodel[0]
    qnodes_ops = {n['name']: n['op'].lower() for n in json.loads(qsymnet.tojson())['nodes']}

    # collect fp32 tensors
    collector = NameCollector()
    run_forward(sym_model, ctx, dataloader, [True], collector)
    tensors = set(collector.names)

    # map tensors to nodes
    tensor_to_node = {}
    nodes = set(nodes_ops.keys())
    for tensor in tensors:
        node = _tensor_to_node(tensor, nodes)
        if node != '':
            tensor_to_node[tensor] = node
        elif tensor in calib_tensors:
            tensor_to_node[tensor] = tensor
    assert set(calib_tensors).issubset(set(tensor_to_node.keys()))
    calib_nodes = [tensor_to_node[ct] for ct in calib_tensors]

    quantizable = {}
    for qsym in qsymnet.get_internals():
        if qnodes_ops[qsym.name] in NULL_OP_NAMES:
            continue
        sym_name, op_type = _dequantize_sym_name(qsym.name, nodes_ops.keys())
        node_name = _tensor_to_node(sym_name, nodes_ops.keys())
        node_name = sym_name if node_name == '' else node_name
        assert qnodes_ops[qsym.name] not in QUANTIZE_OP_NAMES or (op_type == OpType.QUANTIZE), \
            'Quantize node was not recognised properly. Node name: "{}"'.format(node_name)
        if node_name in calib_nodes:
            if op_type == OpType.QUANTIZE:
                quantizable[node_name] = QUANTIZE_OP_NAME
            elif op_type == OpType.QUANTIZED:
                quantizable[node_name] = nodes_ops[node_name]

    quantizable_nodes = [{'name': name, 'type': op} for (name, op) in quantizable.items()]
    op_nodes = {k: v for k, v in nodes_ops.items() if v not in NULL_OP_NAMES}
    return quantizable_nodes, tensor_to_node, op_nodes


def quantize_sym_model(sym_model, ctx, qconfig):
    """Quantizes the symbolic model according to the configuration.

    Args:
        sym_model (tuple): symbol model (symnet, args, auxs).
        qconfig (dict): quantization configuration.

    Returns:
        tuple: Symbol model (symnet, args, auxs) and list of tensors for calibration.
    """

    assert isinstance(sym_model, tuple) and isinstance(sym_model[0], mx.symbol.Symbol)

    symnet, args, auxs = sym_model
    if not check_mx_version('1.7.0'):
        qconfig.pop('quantize_granularity', None)

    arguments = {'sym': symnet, 'offline_params': list(args.keys())}
    arguments.update(qconfig)
    if check_mx_version('2.0.0'):
        arguments['device'] = ctx
    else:
        arguments['ctx'] = ctx
    qsymnet, calib_tensors = mx.contrib.quantization._quantize_symbol(**arguments)
    # args = mx.contrib.quantization._quantize_params(qsymnet, args, {})
    return ((qsymnet, args, auxs), calib_tensors)


def _tensor_to_node(tensor, nodes):
    """Map tensor to one of the nodes. This function assumes, that node tensors 
    (weights, outputs, etc) contain node name in their names.
    """
    assert len(nodes) > 0, '`nodes` cannot be empty'

    PATTERNS = {'',
                '_output[0-9]*$',
                '_[0-9]+$'}
    mapping = []
    for pattern in PATTERNS:
        node = re.sub(pattern, '', tensor)
        if node in nodes and node not in mapping:
            mapping.append(node)
            assert len(mapping) == 1, 'Tensor matched to more than one node. ' \
                'Tensor: {}, matched: {}'.format(tensor, mapping)
    return mapping[0] if len(mapping) > 0 else ''


def _qtensor_to_tensor(qtensor, tensors):
    """Map quantized tensor to its fp32 equivalent. New ops may require updating the patterns.
    Tensors of quantize nodes (which are not present in fp32 models) will be mapped to their input
    nodes.
    """
    assert len(tensors) > 0, '`tensors` cannot be empty'

    PATTERNS = {'quantize': '',
                '_quantize_output0': '',
                '_quantize_0': '',
                '_0_quantize_output0': '_output',
                '_0_quantize_0': '_output',
                '_([0-9]+)_quantize_output0': '_output\g<1>',
                '_([0-9]+)_quantize_0': '_output\g<1>',
                'quantized_': ''}
    mapping = []
    for pattern, repl in PATTERNS.items():
        tensor = re.sub(pattern, repl, qtensor)
        if tensor in tensors and tensor not in mapping:
            mapping.append(tensor)
            assert len(mapping) == 1, \
                'Quantized tensor matched more than one fp32 tensor. ' \
                'Quantized tensor: {}, matched: {}'.format(qtensor, mapping)
    return mapping[0] if len(mapping) > 0 else ''


def run_forward(sym_model, ctx, dataloader, b_filter,
                collector=None, pre_batch=None, post_batch=None):
    """Run forward propagation on the model.

    Args:
        sym_model (tuple): symbol model (symnet, args, auxs).
        dataloader (DataLoaderWrap): data loader.
        b_filter (generator): filter on which batches to run inference on.
        collector (object): collects information during inference.
        pre_batch: function to call prior to batch inference.
        post_batch: function to call after batch inference.

    Returns:
        int: batch count.
    """
    assert isinstance(dataloader, DataLoaderWrap)
    assert collector is None or (hasattr(collector, 'collect_gluon') and
                                 hasattr(collector, 'collect_module'))

    if check_mx_version('2.0.0'):
        sym_block = make_symbol_block(sym_model, ctx, dataloader.input_desc)
        if collector is not None:
            sym_block.register_op_hook(collector.collect_gluon, monitor_all=True)
        return _gluon_forward(sym_block, ctx, dataloader, b_filter, pre_batch, post_batch)
    else:
        mod = make_module(sym_model, ctx, dataloader.input_desc)
        if collector is not None:
            mod._exec_group.execs[0].set_monitor_callback(
                collector.collect_module, monitor_all=True)
    return _module_forward(mod, dataloader, b_filter, pre_batch, post_batch)


def make_symbol_block(sym_model, ctx, input_desc):
    """Convert a symbol model to gluon SymbolBlock.

    Args:
        sym_model (tuple): symbol model (symnet, args, auxs).
        input_desc (list): model input data description.

    Returns:
        mx.gluon.SymbolBlock: SymbolBlock model.
    """

    assert isinstance(sym_model, tuple) and isinstance(sym_model[0], mx.symbol.Symbol)

    symnet, args, auxs = sym_model
    inputs = [mx.sym.var(d.name) for d in input_desc]
    sym_block = mx.gluon.SymbolBlock(symnet, inputs)
    param_dict = args
    param_dict.update(auxs)
    if check_mx_version('2.0.0'):
        sym_block.load_dict(param_dict, cast_dtype=True, dtype_source='saved', allow_missing=True)
    else:
        # params = {'arg:' + name: param for name, param in args.items()}
        # params.update({'aux:' + name: param for name, param in auxs.items()})
        sym_block.collect_params().load_dict(param_dict, ctx=ctx, cast_dtype=True,
                                             dtype_source='saved')
    return sym_block


def _gluon_forward(net, ctx, dataloader, b_filter, pre_batch=None, post_batch=None):
    batch_num = 0
    for run, batch in zip(b_filter, dataloader):
        if not run:
            continue
        batch_num += 1
        batch = ensure_list(batch)
        batch = [ndarray_to_device(d, ctx) for d in batch]
        data = batch[:len(dataloader.input_desc)]
        label = batch[len(dataloader.input_desc):]

        if pre_batch is not None:
            pre_batch(net, (data, label))

        out = net(*data)

        if post_batch is not None:
            post_batch(net, (data, label), out)
    return batch_num


def make_module(sym_model, ctx, input_desc):
    """Convert a symbol model to Module.

    Args:
        sym_model (tuple): symbol model (symnet, args, auxs).
        input_desc (list): model input data description.

    Returns:
        mx.module.Module: Module model.
    """

    assert isinstance(sym_model, tuple) and isinstance(sym_model[0], mx.symbol.Symbol)

    symnet, args, auxs = sym_model
    mod = mx.module.module.Module(symbol=symnet,
                                  data_names=[d.name for d in input_desc],
                                  label_names=None,
                                  context=ctx)
    mod.bind(input_desc, for_training=False)
    mod.set_params(args, auxs, allow_missing=True)
    return mod


def _module_forward(module, dataloader, b_filter, pre_batch=None, post_batch=None):
    batch_num = 0
    for run, batch in zip(b_filter, dataloader):
        if not run:
            continue
        batch_num += 1
        data = batch[:len(dataloader.input_desc)]
        label = batch[len(dataloader.input_desc):]

        if pre_batch is not None:
            pre_batch(module, (data, label))

        module.forward(mx.io.DataBatch(data=data), is_train=False)

        if post_batch is not None:
            post_batch(module, (data, label), module.get_outputs())
    return batch_num


def parse_tune_config(tune_cfg, quantizable_nodes):
    """Convert the strategy config to MXNet quantization config.

    Args:
        tune_cfg (dict): tune config from neural_compressor strategy.
        quantizable_nodes (list): quantizable nodes in the model.

    Returns:
        dict: quantization configuration.
        dict: calibration configuration.
    """
    excluded_symbols = []
    calib_minmax_nodes = set()
    calib_kl_nodes = set()
    amp_excluded_nodes = set()

    for op in quantizable_nodes:
        cfg = tune_cfg['op'][(op['name'], op['type'])]['activation']
        if cfg['dtype'] not in ['bf16']:
            amp_excluded_nodes.add(op['name'])
        if cfg['dtype'] not in ['int8']:
            excluded_symbols.append(op['name'])
            # config for quantize node, that might be added after this node
            # (to quantize its output)
            cfg['algorithm'] = QUANTIZE_DEFAULT_ALGORITHM

        if cfg['algorithm'] == 'kl':
            calib_kl_nodes.add(op['name'])
        else:
            calib_minmax_nodes.add(op['name'])
    assert len(calib_kl_nodes & calib_minmax_nodes) == 0

    quant_cfg = {'excluded_symbols': excluded_symbols,
                 'quantized_dtype': 'auto',
                 'quantize_mode': 'smart'}
    if check_mx_version('1.7.0'):
        quant_cfg['quantize_granularity'] = 'tensor-wise'

    calib_cfg = {'quantized_dtype': quant_cfg['quantized_dtype'],
                 'batches': tune_cfg['calib_iteration'],
                 'calib_mode': 'naive',
                 'calib_kl_nodes': calib_kl_nodes,
                 'calib_minmax_nodes': calib_minmax_nodes}

    amp_cfg = {'target_dtype': 'bfloat16',
               'excluded_sym_names': amp_excluded_nodes}

    return quant_cfg, calib_cfg, amp_cfg


def distribute_calib_tensors(calib_tensors, calib_cfg, tensor_to_node):
    """Distributes the tensors for calibration, depending on the
       algorithm set in the configuration of their nodes.

    Args:
        calib_tensors: tensors to distribute.
        calib_cfg (dict): calibration configuration.
        tensor_to_node (dict): tensor to node mapping.

    Returns:
        tuple: kl tensors and minmax tensors.
    """
    calib_tensors = set(calib_tensors)
    kl_tensors = {}
    minmax_tensors = {}
    for cl in calib_tensors:
        assert cl in tensor_to_node, '`calib_tensors` entry matched no node. Entry: {}'.format(cl)
        node = tensor_to_node[cl]
        if node in calib_cfg['calib_kl_nodes']:
            kl_tensors[cl] = node
        if node in calib_cfg['calib_minmax_nodes']:
            minmax_tensors[cl] = node

    kl_tensors = set(kl_tensors.keys())
    minmax_tensors = set(minmax_tensors.keys())
    assert len(kl_tensors & minmax_tensors) == 0, 'same `calib_tensors` entries matched both kl ' \
        'and minmax nodes. Entries: {}'.format(kl_tensors & minmax_tensors)

    # `rest` are the nodes that require callibration because of some node being excluded
    # for example: input -> quantize -> conv_1 -> pooling -> conv_2
    # when conv_1 is quantized, pooling output does not require callibration
    # when conv_1 is excluded, pooling output requires callibration (as it is input of a quantized
    # node): input -> conv_1 -> pooling -> quantize -> conv_2
    rest = calib_tensors - (kl_tensors | minmax_tensors)
    minmax_tensors |= rest # assign them to the minmax algorithm by default

    return (kl_tensors, minmax_tensors)


def calib_model(qsym_model, calib_data, calib_cfg):
    """Calibrate the quantized symbol model using data gathered by
       the collector.

    Args:
        qsym_model (tuple): quantized symbol model (symnet, args, auxs).
        calib_data (CalibData): data needed for calibration (thresholds).
        calib_cfg (dict): calibration configuration.

    Returns:
        tuple: quantized calibrated symbol model (symnet, args, auxs).
    """

    assert isinstance(qsym_model, tuple) and isinstance(qsym_model[0], mx.symbol.Symbol)

    qsymnet, qargs, auxs = qsym_model
    if check_mx_version('2.0.0'):
        return mx.contrib.quantization.calib_graph(
            qsymnet, qargs, auxs, calib_data, calib_cfg['calib_mode'])
    else:
        return mx.contrib.quantization.calib_graph(
            qsymnet, qargs, auxs, calib_data, calib_cfg['calib_mode'],
            quantized_dtype=calib_cfg['quantized_dtype'])


def amp_convert(sym_model, input_desc, amp_cfg):
    assert check_mx_version('2.0.0'), 'AMP is supported since MXNet 2.0. This error is due to ' \
        'an error in the configuration file.'
    from mxnet import amp

    input_dtypes = {i.name: i.dtype for i in input_desc}
    return amp.convert_model(*sym_model, input_dtypes, **amp_cfg, cast_params_offline=True)


class DataLoaderWrap:
    def __init__(self, dataloader, input_desc):
        self.dataloader = dataloader
        self.input_desc = input_desc
        self._iter = None

    def __iter__(self):
        self._iter = iter(self.dataloader)
        return self

    def __next__(self):
        return next(self._iter)


class DataIterLoader:
    def __init__(self, data_iter):
        self.data_iter = data_iter

    def __iter__(self):
        self.data_iter.reset()
        return self

    def __next__(self):
        batch = self.data_iter.__next__()
        return batch.data + (batch.label if batch.label is not None else [])


class CollectorBase:
    def collect_gluon(self, name, _, arr):
        raise NotImplementedError()

    def collect_module(self, name, arr):
        name = mx.base.py_str(name)
        handle = ctypes.cast(arr, mx.base.NDArrayHandle)
        arr = mx.nd.NDArray(handle, writable=False)
        self.collect_gluon(name, '', arr)

    def pre_batch(self, m, b):
        pass

    def post_batch(self, m, b, o):
        pass


class CalibCollector(CollectorBase):
    """ Collects the calibration thresholds depending on the algorithm set """

    def __init__(self, include_tensors_kl, include_tensors_minmax, num_bins=8001):
        self.min_max_dict = {}
        self.hist_dict = {}
        self.num_bins = num_bins
        self.include_tensors_minmax = include_tensors_minmax
        self.include_tensors_kl = include_tensors_kl

    def collect_gluon(self, name, _, arr):
        if name in self.include_tensors_kl:
            alg = 'kl'
        elif name in self.include_tensors_minmax:
            alg = 'minmax'
        else:
            return

        min_range = arr.min().asscalar()
        max_range = arr.max().asscalar()
        th = max(abs(min_range), abs(max_range))
        # minmax (always)
        if name in self.min_max_dict:
            cur_min_max = self.min_max_dict[name]
            self.min_max_dict[name] = (min(cur_min_max[0], min_range),
                                       max(cur_min_max[1], max_range))
        else:
            self.min_max_dict[name] = (min_range, max_range)

        if alg == 'kl':  # histogram only when kl is specified
            arr = arr.asnumpy()
            if name in self.hist_dict:
                self.hist_dict[name] = self._combine_histogram(self.hist_dict[name], arr,
                                                               min_range, max_range, th)
            else:
                hist, hist_edges = np.histogram(arr, bins=self.num_bins, range=(-th, th))
                self.hist_dict[name] = (hist, hist_edges, min_range, max_range, th)

    @staticmethod
    def _combine_histogram(old_hist, arr, new_min, new_max, new_th):
        if check_mx_version('2.0.0'):
            return mx.contrib.quantization._LayerHistogramCollector.combine_histogram(
                old_hist, arr, new_min, new_max, new_th)
        else:
            return mx.contrib.quantization.combine_histogram(old_hist, arr, new_min,
                                                             new_max, new_th)

    def calc_kl_th_dict(self, quantized_dtype):
        if len(self.hist_dict) > 0:
            if check_mx_version('2.0.0'):
                return mx.contrib.quantization._LayerHistogramCollector.get_optimal_thresholds(
                    self.hist_dict, quantized_dtype)
            else:
                return mx.contrib.quantization._get_optimal_thresholds(
                    self.hist_dict, quantized_dtype)
        return {}


class TensorCollector(CollectorBase):
    """ Tensors collector. Builds up qtensor_to_tensor mapping """

    def __init__(self, include_nodes, qtensor_to_tensor, tensor_to_node):
        self.tensors_dicts = []
        self.include_nodes = dict(include_nodes)
        self.qtensor_to_tensor = qtensor_to_tensor
        self.tensor_to_node = tensor_to_node

        rest = set(self.include_nodes.keys()) - set(self.tensor_to_node.values())
        assert len(rest) == 0, 'Unexpected tensors set to be collected: {}'.format(rest)

    def collect_gluon(self, name, _, arr):
        is_quantized = False
        if name not in self.tensor_to_node:
            if name in self.qtensor_to_tensor:
                name = self.qtensor_to_tensor[name]
            else:
                qname, name = name, _qtensor_to_tensor(name, self.tensor_to_node)
                self.qtensor_to_tensor[qname] = name
            if name == '':
                return
            is_quantized = arr.dtype in QUANTIZATION_DTYPES

        node = self.tensor_to_node[name]
        if node in self.include_nodes:
            op = self.include_nodes[node]
            key = (node, op)
            self.tensors_dicts[-1].setdefault(key, {})[name] = (is_quantized, arr.copy())

    def pre_batch(self, m, b):
        self.tensors_dicts.append({})


class NameCollector(CollectorBase):
    def __init__(self):
        self.names = []

    def collect_gluon(self, name, _, arr):
        self.names.append(name)


class CalibData:
    def __init__(self, cache_kl={}, cache_minmax={}, tensors_kl=[], tensors_minmax=[]):
        self.th_dict = {}
        self.th_dict.update({t: cache_kl[t] for t in tensors_kl})
        self.th_dict.update({t: cache_minmax[t] for t in tensors_minmax})

    # `min_max_dict` is used as a thresholds dictionary when `calib_mode` == 'naive'
    @property
    def min_max_dict(self):
        return self.th_dict

    # for mxnet version >= 2.0.0
    def post_collect(self):
        return self.th_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np
from neural_compressor.utils import logger
from neural_compressor.utils.utility import LazyImport
from neural_compressor.experimental.data.transforms import transform_registry, BaseTransform

tf = LazyImport('tensorflow')
cv2 = LazyImport('cv2')

# BELOW IS TO BE DEPRECATED!
@transform_registry(transform_type="ParseDecodeImagenet", \
                    process="preprocess", framework="tensorflow")
class ParseDecodeImagenetTransform(BaseTransform):

    def __call__(self, sample):
        logger.warning("This transform is going to be deprecated, " \
            "imagenet decoding will be performed automatically from Neural Compressor v1.4.")
        return sample

@transform_registry(transform_type="ResizeCropImagenet", \
                    process="preprocess", framework="tensorflow")
class TensorflowResizeCropImagenetTransform(BaseTransform):
    """Combination of a series of transforms which is applicable to images in Imagenet.

    Args:
        height (int): Height of the result
        width (int): Width of the result
        random_crop (bool, default=False): whether to random crop
        resize_side (int, default=256):desired shape after resize operation
        random_flip_left_right (bool, default=False): whether to random flip left and right
        mean_value (list, default=[0.0,0.0,0.0]):means for each channel
        scale (float, default=1.0):std value

    Returns:
        tuple of processed image and label
    """

    def __init__(self, height, width, random_crop=False, resize_side=256, \
                 resize_method='bilinear', random_flip_left_right=False, \
                 mean_value=[0.0,0.0,0.0], scale=1.0, \
                 data_format='channels_last', subpixels='RGB'):

        self.height = height
        self.width = width
        self.mean_value = mean_value
        self.scale = scale
        self.random_crop = random_crop
        self.random_flip_left_right = random_flip_left_right
        self.resize_side = resize_side
        self.resize_method = resize_method
        self.data_format = data_format
        self.subpixels = subpixels

    # sample is (images, labels)
    def __call__(self, sample):
        image, label = sample
        shape = tf.shape(input=image)
        
        height = tf.cast(shape[0], dtype=tf.float32) \
            if self.data_format=="channels_last" else tf.cast(shape[1], dtype=tf.float32)
        width = tf.cast(shape[1], dtype=tf.float32) \
            if self.data_format=="channels_last" else tf.cast(shape[2], dtype=tf.float32)
        scale = tf.cond(pred=tf.greater(height, width), \
                        true_fn=lambda: self.resize_side / width,
                        false_fn=lambda: self.resize_side / height,)

        scale = tf.cast(scale, dtype=tf.float32)
        new_height = tf.cast(tf.math.rint(height*scale), dtype=tf.int32)
        new_width = tf.cast(tf.math.rint(width*scale), dtype=tf.int32)

        # image = tf.cond(pred=tf.greater(shape[0], shape[1]), \
        #                 false_fn=lambda: tf.image.resize(image, \
        #                     tf.convert_to_tensor(value=[self.resize_side*shape[0]/shape[1], \
        #                         self.resize_side], dtype=tf.int32)),
        #                 true_fn=lambda: tf.image.resize(image, \
        #                     tf.convert_to_tensor(value=[self.resize_side, \
        #                         self.resize_side * shape[1] / shape[0]], dtype=tf.int32)),
        #                )

        if self.subpixels=='BGR' and self.data_format=='channels_first':
            # 'RGB'->'BGR'
            image = tf.cond(tf.equal(tf.rank(image), 3),
                            lambda: tf.experimental.numpy.moveaxis(image[::-1, ...], 0, -1),
                            lambda: tf.experimental.numpy.moveaxis(image[:, ::-1, ...], 1, -1))
        elif self.subpixels=='BGR':
            # 'RGB'->'BGR'
            image = image[..., ::-1]
        image = tf.expand_dims(image, 0)
        image = tf.image.resize(image, [new_height, new_width],
                                method=self.resize_method)
        image = tf.squeeze(image)
        
        shape = tf.shape(input=image)
        if self.random_crop:
            y0 = tf.random.uniform(shape=[], minval=0, maxval=(shape[0] - self.height +1), 
                    dtype=tf.dtypes.int32)
            x0 = tf.random.uniform(shape=[], minval=0, maxval=(shape[1] - self.width +1), 
                    dtype=tf.dtypes.int32)
        else:
            y0 = (shape[0] - self.height) // 2
            x0 = (shape[1] - self.width) // 2

        image = tf.image.crop_to_bounding_box(image, y0, x0, self.height, self.width)
        image.set_shape([self.height, self.width, 3])
        if self.random_flip_left_right:
            image = tf.image.random_flip_left_right(image)
        means = tf.broadcast_to(self.mean_value, tf.shape(input=image))
        image = (image - means) * self.scale
        return (image, label)

@transform_registry(transform_type="BilinearImagenet", \
                    process="preprocess", framework="tensorflow")
class BilinearImagenetTransform(BaseTransform):
    """Combination of a series of transforms which is applicable to images in Imagenet.

    Args:
        height: Height of the result
        width:Width of the result  
        central_fraction(float, default=0.875):fraction of size to crop
        mean_value(list, default=[0.0,0.0,0.0]):means for each channel
        scale(float, default=1.0):std value

    Returns:
        tuple of processed image and label
    """

    def __init__(self, height, width, central_fraction=0.875,
                 mean_value=[0.0,0.0,0.0], scale=1.0):

        self.height = height
        self.width = width
        self.mean_value = mean_value
        self.scale = scale
        self.central_fraction = central_fraction

    # sample is (images, labels)
    def __call__(self, sample):
        image, label = sample
        if image.dtype is not tf.float32:
            image = tf.image.convert_image_dtype(image, dtype=tf.float32)
        # Crop the central region of the image containing 87.5% area of the original image.
        if self.central_fraction:
            image = tf.image.central_crop(image, central_fraction=self.central_fraction)

        if self.height and self.width:
            # Resize the image to the specified height and width.
            image = tf.expand_dims(image, 0)
            image = tf.image.resize(image, [self.height, self.width], \
                                    method=tf.image.ResizeMethod.BILINEAR)
            image = tf.squeeze(image, [0])

        image = tf.subtract(image, 0.5)
        image = tf.multiply(image, 2.0)
        means = tf.broadcast_to(self.mean_value, tf.shape(input=image))
        image = (image - means) * self.scale

        return (image, label)

@transform_registry(transform_type="BilinearImagenet", process="preprocess", \
                    framework="onnxrt_qlinearops, onnxrt_integerops")
class OnnxBilinearImagenetTransform(BaseTransform):
    """Combination of a series of transforms which is applicable to images in Imagenet.

    Args:
        height: Height of the result
        width:Width of the result  
        central_fraction(float, default=0.875):fraction of size to crop
        mean_value(list, default=[0.0,0.0,0.0]):means for each channel
        scale(float, default=1.0):std value

    Returns:
        tuple of processed image and label
    """

    def __init__(self, height, width, central_fraction=0.875,
                 mean_value=[0.0,0.0,0.0], scale=1.0):
        self.height = height
        self.width = width
        self.mean_value = mean_value
        self.scale = scale
        self.central_fraction = central_fraction

    def __call__(self, sample):
        image, label = sample
        if isinstance(image, np.ndarray):
            image = image.astype('float32') / 255. 
        img_shape = image.shape
        depth = img_shape[2]
        img_hd = float(img_shape[0])
        bbox_h_start = int((img_hd - img_hd * self.central_fraction) / 2)
        img_wd = float(img_shape[1])
        bbox_w_start = int((img_wd - img_wd * self.central_fraction) / 2)

        bbox_h_size = img_shape[0] - bbox_h_start * 2
        bbox_w_size = img_shape[1] - bbox_w_start * 2

        image = image[bbox_h_start:bbox_h_start+bbox_h_size, bbox_w_start:bbox_w_start+bbox_w_size]

        if self.height and self.width:
            image = cv2.resize(image, (self.width, self.height), interpolation=cv2.INTER_LINEAR)
        
        image = np.subtract(image, 0.5)
        image = np.multiply(image, 2.0)
        means = np.broadcast_to(self.mean_value, image.shape)
        image = (image - means) * self.scale
        image = image.astype(np.float32)

        return (image, label)

@transform_registry(transform_type="ResizeCropImagenet", process="preprocess", \
                    framework="onnxrt_qlinearops, onnxrt_integerops")
class ONNXResizeCropImagenetTransform(BaseTransform):
    """Combination of a series of transforms which is applicable to images in Imagenet.

    Args:
        height: Height of the result
        width:Width of the result  
        central_fraction(float, default=0.875):fraction of size to crop
        mean_value(list, default=[0.0,0.0,0.0]):means for each channel
        scale(float, default=1.0):std value

    Returns:
        tuple of processed image and label
    """

    def __init__(self, height, width, random_crop=False, resize_side=256, \
                 mean_value=[0.0,0.0,0.0], std_value=[0.229, 0.224, 0.225], \
                 resize_method='bilinear', data_format='channels_last', subpixels='RGB'):

        self.height = height
        self.width = width
        self.mean_value = mean_value
        self.std_value = std_value
        self.random_crop = random_crop
        self.resize_side = resize_side
        self.resize_method = resize_method
        self.data_format = data_format
        self.subpixels = subpixels

    # sample is (images, labels)
    def __call__(self, sample):
        # TODO Support optional resize_method, data_format, subpixels for ONNX
        image, label = sample
        height, width = image.shape[0], image.shape[1]
        scale = self.resize_side / width if height > width else self.resize_side / height
        new_height = int(height*scale)
        new_width = int(width*scale)
        image = cv2.resize(image, (new_height, new_width))
        image = image / 255.
        shape = image.shape
        if self.random_crop:
            y0 = np.random.randint(low=0, high=(shape[0] - self.height +1))
            x0 = np.random.randint(low=0, high=(shape[1] - self.width +1))
        else:
            y0 = (shape[0] - self.height) // 2
            x0 = (shape[1] - self.width) // 2
        if len(image.shape) == 2:
            image = np.array([image])
            image = np.repeat(image, 3, axis=0)
            image = image.transpose(1, 2, 0)
        image = image[y0:y0+self.height, x0:x0+self.width, :]
        image = ((image - self.mean_value)/self.std_value).astype(np.float32)
        return (image.transpose(2, 0, 1), label)

@transform_registry(transform_type="ResizeWithAspectRatio", process="preprocess", \
                    framework="onnxrt_qlinearops, onnxrt_integerops")
class ResizeWithAspectRatio(BaseTransform):
    def __init__(self, height, width, scale=87.5, inter_pol=cv2.INTER_AREA):
        self.height = height
        self.width = width
        self.scale = scale
        self.inter_pol = inter_pol

    def __call__(self, sample):
        (img, label) = sample
        assert len(img.shape) == 3
        height, width, _ = img.shape
        new_height = int(100. * self.height / self.scale) 
        new_width = int(100. * self.width / self.scale)
        if height > width:
            w = new_width
            h = int(new_height * height / width)
        else:
            h = new_height
            w = int(new_width * width / height)
        img = cv2.resize(img, (w, h), interpolation=self.inter_pol)
        return img, label



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Functions for calculating loss, accuracy, and other model metrics.

 BLEU approximation. Source:
     https://github.com/tensorflow/tensor2tensor/blob/master/tensor2tensor/utils/bleu_hook.py
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import math

import numpy as np
import six
from six.moves import xrange  # pylint: disable=redefined-builtin
from neural_compressor.utils.utility import LazyImport
tf = LazyImport('tensorflow')

def _get_ngrams_with_counter(segment, max_order):
    """Extracts all n-grams up to a given maximum order from an input segment.

    Args:
        segment: text segment from which n-grams will be extracted.
        max_order: maximum length in tokens of the n-grams returned by this
            methods.

    Returns:
        The Counter containing all n-grams up to max_order in segment
        with a count of how many times each n-gram occurred.
    """
    ngram_counts = collections.Counter()
    for order in xrange(1, max_order + 1):
        for i in xrange(0, len(segment) - order + 1):
            ngram = tuple(segment[i:i + order])
            ngram_counts[ngram] += 1
    return ngram_counts


def compute_bleu(reference_corpus, translation_corpus, max_order=4,
                 use_bp=True):
    """Computes BLEU score of translated segments against one or more references.

    Args:
        reference_corpus: list of references for each translation. Each
            reference should be tokenized into a list of tokens.
        translation_corpus: list of translations to score. Each translation
            should be tokenized into a list of tokens.
        max_order: Maximum n-gram order to use when computing BLEU score.
        use_bp: boolean, whether to apply brevity penalty.

    Returns:
        BLEU score.
    """
    reference_length = 0
    translation_length = 0
    bp = 1.0
    geo_mean = 0

    matches_by_order = [0] * max_order
    possible_matches_by_order = [0] * max_order
    precisions = []

    for (references, translations) in zip(reference_corpus, translation_corpus):
        reference_length += len(references)
        translation_length += len(translations)
        ref_ngram_counts = _get_ngrams_with_counter(references, max_order)
        translation_ngram_counts = _get_ngrams_with_counter(translations, max_order)

        overlap = dict((ngram,
                        min(count, translation_ngram_counts[ngram]))
                       for ngram, count in ref_ngram_counts.items())

        for ngram in overlap:
            matches_by_order[len(ngram) - 1] += overlap[ngram]
        for ngram in translation_ngram_counts:
            possible_matches_by_order[len(ngram) - 1] += translation_ngram_counts[
                ngram]

    precisions = [0] * max_order
    smooth = 1.0

    for i in xrange(0, max_order):
        if possible_matches_by_order[i] > 0:
            precisions[i] = float(matches_by_order[i]) / possible_matches_by_order[i]
            if matches_by_order[i] > 0:
                precisions[i] = float(matches_by_order[i]) / possible_matches_by_order[
                    i]
            else:
                smooth *= 2
                precisions[i] = 1.0 / (smooth * possible_matches_by_order[i])
        else:
            precisions[i] = 0.0

    if max(precisions) > 0:
        p_log_sum = sum(math.log(p) for p in precisions if p)
        geo_mean = math.exp(p_log_sum / max_order)

    if use_bp:
        ratio = translation_length / reference_length
        bp = math.exp(1 - 1. / ratio) if ratio < 1.0 else 1.0
    bleu = geo_mean * bp
    return np.float32(bleu)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import onnx
import logging
import numpy as np
import copy
from onnx import onnx_pb as onnx_proto
from onnx import TensorProto
from onnx import shape_inference
from onnxruntime.quantization.quant_utils import QuantizedValueType
from onnxruntime.quantization.quant_utils import find_by_name, get_elem_index, get_mul_node, \
                                generate_identified_filename, attribute_to_kwarg, type_to_name
from onnxruntime.quantization.quant_utils import __producer__, __version__, onnx_domain
from onnxruntime import SessionOptions, InferenceSession, GraphOptimizationLevel
from onnxruntime.quantization.quant_utils import QuantizationMode

from neural_compressor.adaptor.ox_utils.registry import CreateQDQQuantizer, \
    CreateOpConverter, CreateCaster
from neural_compressor.adaptor.ox_utils.util import QuantizedValue, QuantizedInitializer, \
    quantize_data_with_scale_zero, quantize_data, dtype_mapping, support_pair, ValueInfo, \
    _get_qrange_for_qType, convert_np_to_float16, cast_tensor, make_quant_node, make_dquant_node
from neural_compressor import options
from neural_compressor.utils.utility import CpuInfo
from neural_compressor.model.onnx_model import ONNXModel

logger = logging.getLogger()


class Quantizer:
    def __init__(self, model, q_config, mode, static, quantization_params,
                 op_types_to_quantize, fallback_list=['fp32']):
        model = onnx.shape_inference.infer_shapes(model)
        self.model = ONNXModel(model)
        self.config = q_config
        self.reduce_range = False if CpuInfo().vnni else True
        self.mode = mode # QuantizationMode.Value
        self.static = static  # use static quantization for inputs.
        self.fuse_dynamic_quant = False
        self.quantization_params = quantization_params
        self.op_types_to_quantize = op_types_to_quantize
        self.fallback_list = fallback_list
        self.new_nodes = []

        self.opset_version = self.check_opset_version()
        self.value_infos = {vi.name: vi for vi in model.graph.value_info}
        self.value_infos.update({ot.name: ot for ot in model.graph.output})
        self.value_infos.update({it.name: it for it in model.graph.input})
        self.replace_input = []
        self.remove_nodes = []
        # List of quantized weights
        self.quantized_value_map = {}
        self.new_value_info = {}

        # QuantizeRange tensor name and zero tensor name for scale and zero point calculation.
        # Used when static is False
        self.fixed_qrange_uint8_name = "fixed_quantization_range_uint8"
        self.fixed_qrange_int8_name = "fixed_quantization_range_int8"
        # For uint8 data-type, to compute zero point, we subtract rmin from 0 (represented by fixed_zero_name tensor)
        self.fixed_zero_name = "fixed_zero"
        # For int8 data-type, zero point is always zero (respresented by fixed_zero_point_name tensor)
        self.fixed_zero_zp_name = "fixed_zero_zp"

        if not self.static:
            self.op_types_to_exclude_output_quantization = op_types_to_quantize
        else:
            self.op_types_to_exclude_output_quantization = [] \
                if not options.onnxrt.qdq_setting.OpTypesToExcludeOutputQuantizatioin \
                else options.onnxrt.qdq_setting.OpTypesToExcludeOutputQuantizatioin

        self.add_qdq_pair_to_weight = options.onnxrt.qdq_setting.AddQDQPairToWeight
        self.dedicated_qdq_pair = options.onnxrt.qdq_setting.DedicatedQDQPair

    def check_opset_version(self):
        ai_onnx_domain = [
            opset for opset in self.model.model.opset_import if not opset.domain \
                                                        or opset.domain == "ai.onnx"]
        if 1 != len(ai_onnx_domain):
            raise ValueError('Failed to find proper ai.onnx domain')
        opset_version = ai_onnx_domain[0].version

        if opset_version > 10:
            self.fuse_dynamic_quant = True
        elif opset_version < 10:
            logger.warning(
                "Warning: The original model opset version is {}, which does not support node \
                fusions. Please update the model to opset >= 11 for better performance."
                .format(opset_version))
            self.model.model.opset_import.remove(ai_onnx_domain[0])
            self.model.model.opset_import.extend([onnx.helper.make_opsetid("", 11)])
            opset_version = 11

        if opset_version < 13 and self.mode == 'qdq':
            logger.warning("Per-Channel support with QDQ format requires onnx opset >= 13," \
                " use per-tensor granularity instead")
        return opset_version

    def should_quantize(self, node):
        if node.name in self.config and self.config[node.name] not in self.fallback_list:
            return True
        elif node.name.split('_quant')[0] in self.config and \
            self.config[node.name.split('_quant')[0]] not in self.fallback_list:
            return True
        else:
            return False

    def quantize_model(self):
        # step 1: insert q-dq, cast-cast pairs
        self.insert_qdq()
 
        # step 2: remove redundant pairs -> qdq model
        self.remove_redundant_pairs()
 
        # step 3: convert q-node-dq to qlinear op if needed
        if self.mode != 'qdq':
            self.convert_qdq_to_operator_oriented()
 
        self.merge_dedicated_qdq_pair() 
 
        self.model.remove_unused_constant()

        self.model.model.producer_name = __producer__
        self.model.model.producer_version = __version__

        return self.model.model

    def merge_dedicated_qdq_pair(self):
        self.remove_nodes = []
        self.replace_input = []
        self.new_nodes = []
        if self.mode == 'qdq' and self.dedicated_qdq_pair:
            for node in self.model.nodes():
                if node.op_type in ['QuantizeLinear']:
                    children = self.model.get_children(node)
                    if len([i for i in children if i.op_type in ['DequantizeLinear']]) < 2:
                        continue
                    for idx, child in enumerate(children):
                        if child.op_type not in ['DequantizeLinear']:
                            continue
                        if self.should_quantize(self.model.get_children(child)[0]):
                            inputs = [self.model.get_parents(node)[0].output[0],
                                      node.input[1], node.input[2]]
                            self.new_nodes.append(onnx.helper.make_node(
                                "QuantizeLinear",
                                inputs,
                                [node.output[0] + '_' + str(idx)],
                                node.name + '_' + str(idx)))
                            self.replace_input.append([child, node.output[0], 
                                                node.output[0] + '_' + str(idx)])
                        else:
                            self.remove_nodes.append(child)
                            self.replace_input.append([self.model.get_children(child)[0],
                                                child.output[0], node.input[0]])
                    self.remove_nodes.append(node)
        elif self.mode != 'qdq' or not self.dedicated_qdq_pair:
            target_type = ['QuantizeLinear', 'DequantizeLinear']
            for op_type in target_type:
                for node in self.model.nodes():
                    children = self.model.get_children(node)
                    dq_nodes = [i for i in children if i.op_type == op_type]
                    if len(dq_nodes) < 2 or node.op_type in ['Split']:
                        continue
                    datas = []
                    for n in dq_nodes:
                        datas.append([onnx.numpy_helper.to_array(
                                          find_by_name(n.input[1], self.model.initializer())), 
                                      onnx.numpy_helper.to_array(
                                          find_by_name(n.input[2], self.model.initializer()))])
                    for idx, data in enumerate(datas):
                        repeaded_id = [i for i, item in enumerate(datas[idx:]) if item == data]
                        for i in repeaded_id[1:]:
                            self.remove_nodes.append(dq_nodes[i])
                            self.replace_input.append([self.model.get_children(dq_nodes[i])[0],
                                                       dq_nodes[i].output[0], 
                                                       dq_nodes[idx].output[0]])
        self.model.remove_nodes(self.remove_nodes)
        self.model.graph().node.extend(self.new_nodes)
        for node, old_input_name, new_input_name in self.replace_input:
            self.model.replace_node_input(node, old_input_name, new_input_name)
        self.model.update()

    def should_cast(self, node):
        if node.name in self.config and self.config[node.name] != 'fp32': # pragma: no cover
            return True
        else:
            return False

    def insert_qdq(self):
        for node in self.model.nodes():
            if self.should_quantize(node):
                op_quantizer = CreateQDQQuantizer(self, node)
                op_quantizer.quantize()
            elif self.should_cast(node): # pragma: no cover
                op_caster = CreateCaster(self, node)
                op_caster.cast()
        self.model.graph().node.extend(self.new_nodes)
        self.model.remove_nodes(self.remove_nodes)

        for node, old_input_name, new_input_name in self.replace_input:
            self.model.replace_node_input(node, old_input_name, new_input_name)
        self.model.update()
 
    def should_convert(self, node):
        name = node.name.split('_quant')[0]
        if name in self.config and self.config[name] not in self.fallback_list:
            return True
        else:
            return False

    def convert_qdq_to_operator_oriented(self):
        self.new_nodes = []
        self.remove_nodes = []
        self.replace_input = []
        for node in self.model.nodes():
            if node.op_type not in ['QuantizeLinear', 'DequantizeLinear'] and \
                self.should_convert(node):
                op_converter = CreateOpConverter(self, node)
                op_converter.convert()
        self.model.graph().node.extend(self.new_nodes)
        self.model.remove_nodes(self.remove_nodes)
        for node, old_input_name, new_input_name in self.replace_input:
            self.model.replace_node_input(node, old_input_name, new_input_name)
        self.model.update()

    def remove_redundant_pairs(self):
        self.remove_nodes = []
        self.replace_input = []
        pairs = [['QuantizeLinear', 'DequantizeLinear'], 
                 ['Cast', 'Cast'],
                 ]
        def dfs(match_nodes, node, pattern):
            if len(pattern) == 0:
                return
            start_id = 0
            end_id = len(pattern) - 1

            if node.op_type == pattern[end_id]:
                match_nodes.append(node)
            else:
                return

            if start_id == end_id:
                if all([i.op_type in ['QuantizeLinear', 'DequantizeLinear'] \
                    for i in match_nodes]):
                    pair = [str(find_by_name(i.input[2], self.model.initializer()).data_type) \
                        for i in match_nodes[::-1]]
                    if ' '.join(pair) in support_pair and support_pair[' '.join(pair)]:
                        self.replace_input.append([
                            self.model.get_children(match_nodes[1])[0],
                            match_nodes[1].output[0], 
                            match_nodes[0].input[0]])
 
                        self.remove_nodes.append(match_nodes[1])
                        if all([i.op_type in ['QuantizeLinear', 'DequantizeLinear'] \
                            for i in self.model.get_children(match_nodes[0])]):
                            self.remove_nodes.append(match_nodes[0])
                else:
                    parent = self.model.get_parents(match_nodes[0])[0]
                    children = self.model.get_children(match_nodes[1])
                    input_dtype = '1' # float32
                    output_dtype = '1' # 'float32'
                    for inp in parent.input:
                        if inp in self.new_value_info:
                            input_dtype = str(self.new_value_info[inp].new_dtype)
                            break
                    for child in children:
                        outs = [out for out in child.output if out in self.new_value_info]
                        if len(outs) > 0:
                            output_dtype = str(self.new_value_info[outs[0]].new_dtype)
                            break
                    if len(outs) == 0 or all([not self.should_convert(i) for i in children]):
                        return
                    if input_dtype == str(match_nodes[1].attribute[0].i) and \
                        output_dtype == str(match_nodes[0].attribute[0].i) and \
                        ' '.join((output_dtype, input_dtype)) in support_pair and \
                        support_pair[' '.join((output_dtype, input_dtype))]:
                        if match_nodes[0] not in self.remove_nodes and \
                            all([i.op_type == 'Cast' and str(i.attribute[0].i) == input_dtype \
                            for i in self.model.get_children(match_nodes[0])]):
                            self.remove_nodes.append(match_nodes[0])
                        if match_nodes[1] not in self.remove_nodes:
                            self.remove_nodes.append(match_nodes[1])
                        for child in children:
                            self.replace_input.append([
                                find_by_name(child.name, self.model.model.graph.node),
                                match_nodes[1].output[0], match_nodes[0].input[0]])
                return

            children = self.model.get_children(node)
            for child in children:
                dfs(copy.deepcopy(match_nodes), child, pattern[:end_id])

        for n in self.model.nodes():
            matched = [i for i in pairs if n.op_type == i[-1]]
            if len(matched) > 0:
                for match_pair in matched:
                    visited_op = []
                    dfs(visited_op, n, match_pair)
        self.model.remove_nodes(self.remove_nodes)
        for node, old_input_name, new_input_name in self.replace_input:
            self.model.replace_node_input(node, old_input_name, new_input_name)
        self.model.update()

    def dtype_cast(self, node, cfg, keep_io_types=True): # pragma: no cover
        min_positive_val = 1e-7
        max_finite_val = 1e4
        for idx, tensor_name in enumerate(node.input):
            initializer = find_by_name(tensor_name, self.model.initializer())
            if initializer is not None:
                if initializer.data_type != onnx_proto.TensorProto.FLOAT: 
                    continue
                cast_tensor(initializer, cfg, min_positive_val, max_finite_val)
                self.new_value_info[tensor_name] = ValueInfo(tensor_name,
                                                         TensorProto.FLOAT, dtype_mapping[cfg])
            else:
                if tensor_name in self.value_infos and \
                    self.value_infos[tensor_name].type.HasField('tensor_type') and \
                    self.value_infos[tensor_name].type.tensor_type.elem_type != TensorProto.FLOAT:
                    continue 
                name = node.name + '_input_cast' + str(idx)
                self.new_nodes.append(onnx.helper.make_node(
                        'Cast', [tensor_name], [name], to=dtype_mapping[cfg], name=name))
                node.input[idx] = name
                self.new_value_info[name] = ValueInfo(tensor_name,
                                                             TensorProto.FLOAT, dtype_mapping[cfg])
        if all([i not in self.new_value_info for i in node.input]):
            return

        for idx, tensor_name in enumerate(node.output):
            if tensor_name in self.value_infos and \
                self.value_infos[tensor_name].type.HasField('tensor_type') and \
                self.value_infos[tensor_name].type.tensor_type.elem_type != TensorProto.FLOAT:
                continue 
            node.output[idx] = tensor_name + "_to_cast_" + str(idx)
            name = node.name + '_output_cast' + str(idx)
            self.new_nodes.append(onnx.helper.make_node(
                    'Cast', [node.output[idx]], [tensor_name], to=1, name=name))
            self.new_value_info[node.output[idx]] = ValueInfo(tensor_name,
                                                         dtype_mapping[cfg], TensorProto.FLOAT)

    def quantize_outputs(self, node, initializer_use_weight_qType=True, direct_int8=False):
        if not self.static:
            return
        for idx, tensor_name in enumerate(node.output):
            if tensor_name in self.value_infos and \
                self.value_infos[tensor_name].type.HasField('tensor_type') and \
                self.value_infos[tensor_name].type.tensor_type.elem_type != TensorProto.FLOAT:
                return
            data_found = False
            if direct_int8:
                if node.input[0] in self.quantized_value_map:
                    scale_name = self.quantized_value_map[node.input[0]].scale_name
                    zp_name = self.quantized_value_map[node.input[0]].zp_name
                    data_found = True
                elif node.input[0] in self.quantization_params:
                    data_found, scale_name, zp_name, _, _ = \
                        self._get_quantization_params(node.input[0])
            else:
                if tensor_name in self.quantized_value_map:
                    scale_name = self.quantized_value_map[tensor_name].scale_name
                    zp_name = self.quantized_value_map[tensor_name].zp_name
                    data_found = True
                elif tensor_name in self.quantization_params:
                    data_found, scale_name, zp_name, _, _ = \
                        self._get_quantization_params(tensor_name)
 
            if data_found == False:
                raise ValueError(
                    "Quantization parameters are not specified for param {}."
                    "In static mode quantization params for inputs and outputs \
                    of nodes to be quantized are required.".format(tensor_name))

            node.output[idx] = tensor_name + "_QuantizeInput"
            q_input = node.output[idx]
            q_output = tensor_name + "_quantized"
            dq_input = q_output
            dq_output = tensor_name
            quant_node_name = tensor_name + "_" + node.name + "_QuantizeLinear"
            dequant_node_name = tensor_name + "_" + node.name + "_DequantizeLinear"
            qlinear_node = make_quant_node(quant_node_name,
                                           [q_input, scale_name, zp_name], [q_output])
            dequant_node = make_dquant_node(dequant_node_name,
                                            [dq_input, scale_name, zp_name], [dq_output])
            self.new_nodes.extend([qlinear_node, dequant_node])
            quantized_value = QuantizedValue(tensor_name, dq_output,
                                             scale_name,
                                             zp_name, 
                                             QuantizedValueType.Input)
            if tensor_name not in self.quantized_value_map:
                self.quantized_value_map[tensor_name] = quantized_value
 
    def quantize_inputs(self, node, indices=None, 
            initializer_use_weight_qType=True, direct_int8=False):
        # Quantize the input
        for idx, tensor_name in enumerate(node.input):
            if indices and idx not in indices:
                continue
            initializer = find_by_name(tensor_name, self.model.initializer())
            if initializer is not None:
                if initializer.data_type != onnx_proto.TensorProto.FLOAT:
                    return
                if node.op_type not in self.op_types_to_quantize:
                    dtype = onnx_proto.TensorProto.INT8 if initializer_use_weight_qType \
                        else onnx_proto.TensorProto.UINT8
                    scheme = 'sym' if initializer_use_weight_qType else 'asym'
                else:
                    dtype = self.config[node.name]['weight']['dtype'] if \
                        initializer_use_weight_qType else \
                        self.config[node.name]['activation']['dtype']
                    scheme = self.config[node.name]['weight']['scheme'] if \
                        initializer_use_weight_qType else \
                        self.config[node.name]['activation']['scheme']
                if self.add_qdq_pair_to_weight and self.mode == 'qdq':
                    weight = self._get_quantized_weight(initializer, dtype, scheme)
                    self._update_weight(weight)
                    q_weight_name = weight.name + "_quantized"
                    zp_name = weight.name + "_zero_point"
                    scale_name = weight.name + "_scale"
                    qlinear_node = make_quant_node(tensor_name + "_QuantizeLinear",
                        [tensor_name, scale_name, zp_name], [tensor_name + "_quantized"])
                    dequant_node = make_dquant_node(tensor_name + "_DequantizeLinear",
                        [tensor_name + "_quantized", scale_name, zp_name],
                        [tensor_name + "_dequantized"])
                    self.replace_input.append([node, tensor_name, dequant_node.output[0]])
                    self.new_nodes.extend([qlinear_node, dequant_node])
                    quantized_value = QuantizedValue(weight.name, q_weight_name,
                                                     scale_name,
                                                     zp_name, 
                                                     QuantizedValueType.Initializer,
                                                     None, dtype)
                    if weight.name not in self.quantized_value_map:
                        self.quantized_value_map[weight.name] = quantized_value
                else:
                    weight = self._get_quantized_weight(initializer, dtype, scheme)
                    self._update_weight(weight)
                    q_weight_name = weight.name + "_quantized"
                    zp_name = weight.name + "_zero_point"
                    scale_name = weight.name + "_scale"
 
                    inputs = [q_weight_name, scale_name, zp_name]
                    output_name = tensor_name + '_DequantizeLinear'
                    dequant_node = onnx.helper.make_node("DequantizeLinear", inputs,
                        [tensor_name + '_dequantized'], tensor_name + '_DequantizeLinear')
                    self.new_nodes.append(dequant_node)
                    self.replace_input.append([node, tensor_name, dequant_node.output[0]])
                    quantized_value = QuantizedValue(weight.name, q_weight_name,
                                                     scale_name,
                                                     zp_name, 
                                                     QuantizedValueType.Initializer,
                                                     None, dtype)
                    if weight.name not in self.quantized_value_map:
                        self.quantized_value_map[weight.name] = quantized_value
            else:
                if tensor_name in self.value_infos and \
                    self.value_infos[tensor_name].type.HasField('tensor_type') and \
                    self.value_infos[tensor_name].type.tensor_type.elem_type != TensorProto.FLOAT:
                    return

                if tensor_name in self.quantized_value_map:
                    scale_name = self.quantized_value_map[tensor_name].scale_name
                    zp_name = self.quantized_value_map[tensor_name].zp_name
                    data_found = True
                else:
                    data_found, scale_name, zp_name, _, _ = \
                        self._get_quantization_params(tensor_name)
 
                if self.static:
                    if data_found == False:
                        raise ValueError(
                            "Quantization parameters are not specified for param {}."
                            "In static mode quantization params for inputs and outputs \
                            of nodes to be quantized are required.".format(tensor_name))
                    if direct_int8:
                        parent = self.model.get_parents(node)[0]
                        if not parent.output[0].endswith('_QuantizeInput'):
                            return
                    q_input = tensor_name
                    q_output = tensor_name + "_" + node.name + "_QuantizeLinear" if \
                        tensor_name not in self.model.input() else tensor_name + "_quantized"
                    dq_input = q_output
                    dq_output = tensor_name + "_" + node.name + "_dequantized" if \
                        tensor_name not in self.model.input() else tensor_name + "_dequantized"
                    self.replace_input.append([node, tensor_name, dq_output])

                    quant_node_name = tensor_name + "_" + node.name + "_QuantizeLinear"
                    dequant_node_name = tensor_name + "_" + node.name + "_DequantizeLinear"
                    qlinear_node = make_quant_node(
                        quant_node_name, [q_input, scale_name, zp_name], [q_output])
                    dequant_node = make_dquant_node(
                        dequant_node_name, [dq_input, scale_name, zp_name], [dq_output])
                    self.new_nodes.extend([qlinear_node, dequant_node])
                    quantized_value = QuantizedValue(tensor_name, dq_output,
                                                     scale_name,
                                                     zp_name, 
                                                     QuantizedValueType.Input)
                    if tensor_name not in self.quantized_value_map:
                        self.quantized_value_map[tensor_name] = quantized_value
                else:
                    qlinear_node = self.model.find_node_by_name(tensor_name + "_QuantizeLinear",
                                                                self.new_nodes,
                                                                self.model.graph())
                    if qlinear_node is None:
                        if data_found == True:
                            qlinear_node = make_quant_node(tensor_name + "_QuantizeLinear",
                                [tensor_name, scale_name, zp_name], [tensor_name + "_quantized"])
                        else:
                            if self.fuse_dynamic_quant and \
                                self.config[node.name]['activation']['dtype'] == \
                                    onnx_proto.TensorProto.UINT8 and \
                                self.config[node.name]['activation']['scheme'] == 'asym':
                                scale_name = tensor_name + "_scale"
                                zeropoint_name = tensor_name + "_zero_point"
                                qlinear_node = onnx.helper.make_node("DynamicQuantizeLinear", 
                                    [tensor_name],
                                    [tensor_name + "_quantized", scale_name, zeropoint_name],
                                    tensor_name + "_QuantizeLinear")
                            else:
                                scale_name, zp_name, _, _ = \
                                    self._get_dynamic_input_quantization_params(
                                    tensor_name, self.config[node.name]['activation']['dtype'])
                                qlinear_node = make_quant_node(tensor_name + "_QuantizeLinear",
                                                            [tensor_name, scale_name, zp_name], 
                                                            [tensor_name + "_quantized"])
                        if qlinear_node not in self.new_nodes:
                            self.new_nodes.append(qlinear_node)
                        self.quantized_value_map[tensor_name] = QuantizedValue(
                            tensor_name, 
                            tensor_name + "_quantized", 
                            scale_name, 
                            zp_name, 
                            self.config[node.name]['activation']['dtype'])                        
                    self.replace_input.append([node, tensor_name, tensor_name + "_quantized"])
 
    def quantize_bias_tensor(self, node):
        input_name, weight_name, bias_name = node.input
        if self.quantization_params is None or input_name not in self.quantization_params and \
            input_name not in self.quantized_value_map:
            self._dynamic_quantize_bias(input_name, weight_name + '_scale', bias_name,
                bias_name + "_quantized")
        else:
            _, quant_value = self.quantize_bias(bias_name, input_name, weight_name)
            self.model.remove_initializer(find_by_name(bias_name, self.model.initializer()))
            inputs = [quant_value.q_name, quant_value.scale_name, quant_value.zp_name]
            dequant_node = onnx.helper.make_node("DequantizeLinear", inputs, 
                [bias_name + '_dequantized'], bias_name + '_DequantizeLinear')
            self.new_nodes.append(dequant_node)
            self.replace_input.append([find_by_name(node.name, self.model.nodes()), 
                bias_name, bias_name + '_dequantized'])

    def quantize_bias(self, bias_name, input_name, weight_name, new_node_list=[]):
        '''
        Quantized the bias. Zero Point == 0 and Scale == Input_Scale * Weight_Scale
        '''
        # get scale for weight
        weight_scale_initializer = find_by_name(weight_name + '_scale', self.model.initializer())
        weight_scale = self.tensor_proto_to_array(weight_scale_initializer)

        # get bias
        bias_initializer = find_by_name(bias_name, self.model.initializer())
        bias_data = self.tensor_proto_to_array(bias_initializer)
        quantized_bias_name = bias_name + "_quantized"

        if input_name in self.quantized_value_map:
            input_scale_name = self.quantized_value_map[input_name].scale_name
        elif input_name in self.quantization_params:
            _, input_scale_name, _, _, _ = self._get_quantization_params(input_name)
        else:
            raise ValueError("Expected {} to be in quantized value map \
                              for static quantization".format(input_name))
        inputscale_initializer = find_by_name(input_scale_name, self.model.initializer())
        input_scale = self.tensor_proto_to_array(inputscale_initializer)

        # calcuate scale for bias

        bias_scale = input_scale * weight_scale

        # quantize bias
        quantized_data = (np.asarray(bias_data) / bias_scale).round().astype(np.int32)

        # update bias initializer
        bias_np_data = np.asarray(quantized_data, dtype=np.int32).reshape(\
                       bias_initializer.dims)
        packed_bias_initializer = onnx.numpy_helper.from_array(bias_np_data, 
                                                               quantized_bias_name)
        self.model.initializer().extend([packed_bias_initializer])

        # update scale initializer
        quantized_bias_scale_name = bias_name + "_scale"
        bias_scale_data = np.asarray(bias_scale, dtype=np.float32).reshape(-1)
        packed_bias_scale_initializer = onnx.numpy_helper.from_array(bias_scale_data,
                                                         quantized_bias_scale_name)
        self.model.initializer().extend([packed_bias_scale_initializer])

        # update zero initializer
        quantized_bias_zp_name = bias_name + "_zero_point"
        bias_zp_data = np.zeros(bias_scale.shape, dtype=np.int32).reshape(-1)
        packed_bias_zp_initializer = onnx.numpy_helper.from_array(
            bias_zp_data, quantized_bias_zp_name)
        self.model.initializer().extend([packed_bias_zp_initializer])

        # log entries for this quantized bias value
        quantized_bias_entry = QuantizedInitializer(bias_name,
                                                    bias_initializer, [0], [0], [0], 
                                                    [bias_scale],
                                                    bias_data,
                                                    quantized_data,
                                                    qType=onnx_proto.TensorProto.INT32)

        quantized_value = QuantizedValue(bias_name, quantized_bias_name, 
                                         quantized_bias_scale_name,
                                         quantized_bias_zp_name, 
                                         QuantizedValueType.Initializer,
                                         None, onnx_proto.TensorProto.INT32)
        return quantized_bias_name, quantized_value

    def _dynamic_quantize_bias(self, input_name, weight_scale_name, \
        bias_name, quantized_bias_name):
        '''
        Adds series of nodes required to quantize the bias dynamically.
            parameter input_name: Input name
            parameter weight_scale_name: Weight scale.
            parameter bias_scale_name: Bias to quantize.
            parameter quantied_bias_name: Output name to use for quantized bias.
        '''
        qType = onnx_proto.TensorProto.INT32
        input_scale_name = input_name + "_scale"
        bias_scale_node = onnx.helper.make_node("Mul",
                                                [input_scale_name, weight_scale_name],
                                                [bias_name + "_scale"],
                                                bias_name + "_scale_node")
        self.new_nodes.append(bias_scale_node)

        quantize_bias_node = onnx.helper.make_node("Div", [bias_name, bias_scale_node.output[0]],
                                                   [bias_name + "_tmp_quant:0"],
                                                   bias_name + "_tmp_qaunt")
        self.new_nodes.append(quantize_bias_node)

        bias_rounded_node = onnx.helper.make_node("Floor", quantize_bias_node.output,
                                                  [bias_name + "_quant_rounded:0"],
                                                  bias_name + "_quant_rounded")
        self.new_nodes.append(bias_rounded_node)

        bias_cast_node = onnx.helper.make_node("Cast",
                                               bias_rounded_node.output, [quantized_bias_name],
                                               quantized_bias_name + "_node",
                                               to=qType)
        self.new_nodes.append(bias_cast_node)
        return

    def quantize_weights_per_channel(self, node, indices, weight_qType, scheme, axis):
        if self.opset_version < 13 and self.mode == 'qdq':
            self.quantize_inputs(node, indices)
            return

        for idx, weight_name in enumerate(node.input):
            if idx not in indices:
                continue

            if self.add_qdq_pair_to_weight and self.mode == 'qdq':
                q_name, zp_name, scale_name = self.quantize_weight_per_channel(weight_name, 
                                                                               weight_qType,
                                                                               scheme,
                                                                               axis) 
                qlinear_node = make_quant_node(weight_name + "_QuantizeLinear",
                        [weight_name, scale_name, zp_name], [weight_name + "_quantized"])
                dequant_node = make_dquant_node(weight_name + "_DequantizeLinear",
                            [weight_name + "_QuantizeLinear", scale_name, zp_name], 
                            [weight_name + "_dequantized"])
                self.replace_input.append([node, weight_name, dequant_node.output[0]])
                self.new_nodes.extend([qlinear_node, dequant_node])
            else:
                q_name, zp_name, scale_name = self.quantize_weight_per_channel(weight_name, 
                                                                               weight_qType,
                                                                               scheme,
                                                                               axis)
                inputs = [q_name, scale_name, zp_name]
                dequant_node = make_dquant_node(weight_name + '_DequantizeLinear',
                    [q_name, scale_name, zp_name], [weight_name + "_dequantized"])
                self.new_nodes.append(dequant_node)

                # Replace weight_name with output of DequantizeLinear
                self.replace_input.append([node, weight_name, dequant_node.output[0]])

    def quantize_weight_per_channel(self, weight_name, weight_qType, scheme, channel_axis):
        initializer = find_by_name(weight_name, self.model.initializer())
        if initializer is None:
            raise ValueError("{} is not an initializer", weight_name)

        if initializer.name not in self.quantized_value_map:
            weights = self.tensor_proto_to_array(initializer)
            channel_count = weights.shape[channel_axis]
            rmin_list = []
            rmax_list = []
            zero_point_list = []
            scale_list = []
            quantized_per_channel_data_list = []
            for i in range(channel_count):
                per_channel_data = weights.take(i, channel_axis)
                rmin, rmax, zero_point, scale, quantized_per_channel_data = quantize_data(
                    per_channel_data.flatten().tolist(), _get_qrange_for_qType(weight_qType, 
                    self.reduce_range), weight_qType, scheme)
                rmin_list.append(rmin)
                rmax_list.append(rmax)
                zero_point_list.append(zero_point)
                scale_list.append(scale)
                quantized_per_channel_data_list.append(quantized_per_channel_data)

            # combine per_channel_data into one
            reshape_dims = list(weights.shape)  # deep copy
            reshape_dims[channel_axis] = 1  # only one per channel for reshape
            quantized_weights = np.asarray(quantized_per_channel_data_list[0]).reshape(reshape_dims)
            for i in range(1, len(quantized_per_channel_data_list)):
                channel_weights = np.asarray(quantized_per_channel_data_list[i]).reshape(reshape_dims)
                quantized_weights = np.concatenate((quantized_weights, channel_weights), channel_axis)

            weight = QuantizedInitializer(initializer.name, initializer, rmin_list, rmax_list, 
                                          zero_point_list, scale_list,
                                          weights,
                                          quantized_weights.flatten().tolist(), 
                                          channel_axis, weight_qType)

            self._update_weight(weight)
            quantized_value = QuantizedValue(weight.name, weight.name + "_quantized",
                                             weight.name + "_scale",
                                             weight.name + "_zero_point",
                                             QuantizedValueType.Initializer,
                                             None, weight_qType)
            self.quantized_value_map[weight.name] = quantized_value
            
        return (initializer.name + "_quantized", initializer.name + "_zero_point", 
                initializer.name + "_scale")

    def _update_weight(self, weight):
        '''
            Given a weight object, update the graph by doing the following:
             - remove old initializer, update new initializers for 
               quantized weight, zero point, and scale
             - remove old weight input, update with new inputs for 
               quantized weight, zero point, and scale
            This function does NOT update the nodes in the graph, just initializers and inputs
        '''
        if weight.name in self.quantized_value_map:
            return
        packed_weight_name = weight.name + "_quantized"
        scale_name = weight.name + "_scale"
        zero_point_name = weight.name + "_zero_point"
        
        # Update packed weight, zero point, and scale initializers
        packed_weight_np_data = np.asarray(weight.quantized_data,
                                           dtype=onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[weight.qType]
                                               ).reshape(weight.initializer.dims)
        packed_weight_initializer = onnx.numpy_helper.from_array(packed_weight_np_data,\
                                                packed_weight_name)

        if not self.add_qdq_pair_to_weight or self.mode != 'qdq':
            self.model.initializer().append(packed_weight_initializer)
        if weight.axis is not None:
            zero_scale_shape = [weight.initializer.dims[weight.axis]]
        else:  # scale and zero point must be scalar
            zero_scale_shape = []
        zero_point_type = weight.qType
        scale_initializer = onnx.helper.make_tensor(scale_name, onnx_proto.TensorProto.FLOAT, 
                                                    zero_scale_shape, weight.scales)
        zero_initializer = onnx.helper.make_tensor(zero_point_name, zero_point_type, 
                                                    zero_scale_shape, weight.zero_points)

        self.model.initializer().extend([scale_initializer, zero_initializer])

    @staticmethod
    def tensor_proto_to_array(initializer):
        if initializer.data_type == onnx_proto.TensorProto.FLOAT:
            weights = onnx.numpy_helper.to_array(initializer)
        else:
            raise ValueError('Only float type quantization is supported. \
               Weights {} is {}. '.format(initializer.name, type_to_name[initializer.data_type]))
        return weights

    def _get_quantization_params(self, param_name):
        '''
        Create initializers and inputs in the graph for zero point and scale of output.
        Zero point and scale values are obtained from self.quantization_params if specified.
            parameter param_name: Name of the quantization parameter.
            return: result, scale_name, zero_point_name, scale_shape, zero_point_shape.
        '''
        if self.quantization_params is None or param_name not in self.quantization_params:
            return False, "", "", "", ""

        params = self.quantization_params[param_name]
        if params is None or len(params) != 2:
            raise ValueError("Quantization parameters should contain zero point and scale. "
                             "Specified values for output {}: {}".format(param_name, params))

        zero_point_values = [params[0].item()]
        zero_point_shape = []
        zero_point_name = param_name + "_zero_point"
        zero_point_type = onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[params[0].dtype]

        scale_values = [params[1].item()]
        scale_shape = []
        scale_name = param_name + "_scale"

        # Add initializers
        init_zp = onnx.helper.make_tensor(zero_point_name, zero_point_type, 
                                          zero_point_shape, zero_point_values)
        self.model.add_initializer(init_zp)
        init_scale = onnx.helper.make_tensor(scale_name, onnx_proto.TensorProto.FLOAT, 
                                             scale_shape, scale_values)
        self.model.add_initializer(init_scale)

        return True, scale_name, zero_point_name, scale_shape, zero_point_shape

    def _get_quantized_weight(self, initializer, qType, scheme):
        '''
            :param initializer: TensorProto initializer
            :param scheme: sym or asym quantization.
            :param qType: type to quantize to
            :return: Weight class with quantization information
        '''
        if initializer.name in self.quantized_value_map:
            return self.quantized_value_map[initializer.name]
        weights_data = self.tensor_proto_to_array(initializer)
        rmin, rmax, zero_point, scale, quantized_weights_data = quantize_data(
            weights_data.flatten().tolist(), _get_qrange_for_qType(qType, \
            self.reduce_range), qType, scheme)
        weight = QuantizedInitializer(initializer.name,
                                      initializer, [rmin], [rmax], [zero_point], [scale],
                                      weights_data,
                                      quantized_weights_data,
                                      axis=None,
                                      qType=qType)

        return weight

    def get_bias_add_nodes(self, node, weight_name, last_output, quantized_bias_name):
        '''
        Given a node, this function handles bias add by
            adding a "reshape" node on bias and an "add" node
            parameter node: current node (Conv)
            parameter last_output: output of previous node (input to bias add)
            return: the name of output
        '''
        # Add tensors for the shape to be reshaped to
        weight = find_by_name(weight_name, self.model.initializer())
        if weight is None:
            raise ValueError("Expected {} to be an initializer".format(node.input[1]))

        # Add reshape for correct broadcase
        reshape_input_data = quantized_bias_name
        reshape_input_shape = quantized_bias_name + "_reshape_shape"
        reshape_input = [reshape_input_data, reshape_input_shape]
        reshape_shape = np.ones((len(weight.dims)), dtype=np.int64)
        reshape_shape[1] = -1
        init_shape = onnx.helper.make_tensor(reshape_input_shape, onnx_proto.TensorProto.INT64,
            [len(weight.dims)], reshape_shape)
        self.model.add_initializer(init_shape)

        reshape_op_output = node.output[0] + "_reshape"
        reshape_node = onnx.helper.make_node("Reshape", reshape_input, [reshape_op_output],
            quantized_bias_name + "reshape")
        self.new_nodes.append(reshape_node)

        # Add an Add operation for bias
        bias_add_input = [last_output]
        bias_add_input.append(reshape_op_output)
        add_node_output = node.output[0] + "_bias_add"
        add_node = onnx.helper.make_node("Add", bias_add_input, [add_node_output],
            quantized_bias_name + "bias_add")
        self.new_nodes.append(add_node)
        return add_node_output

    def is_valid_quantize_weight(self, weight_name):
        weight = find_by_name(weight_name, self.model.initializer())
        if weight is not None:
            return weight.data_type == onnx_proto.TensorProto.FLOAT
        else:
            return weight_name in self.quantized_value_map

    def dequantize_tensor(self, node, value_name):
        if value_name in self.quantized_value_map:
            quantized_value = self.quantized_value_map[value_name]
            dqlinear_name = value_name + "_DequantizeLinear"
            dqlinear_inputs = [value_name + '_quantized',
                               quantized_value.scale_name,
                               quantized_value.zp_name]
            dequantize_node = onnx.helper.make_node("DequantizeLinear",
                                                    dqlinear_inputs,
                                                    [value_name],
                                                    dqlinear_name)
            if dequantize_node not in self.new_nodes:
                self.new_nodes.append(dequantize_node)
        else: # pragma: no cover
            data_found, scale_name, zp_name, _, _ = self._get_quantization_params(value_name)
            if self.static:
               if data_found == False:
                   raise ValueError(
                       "Quantization parameters are not specified for param {}."
                       "In static mode quantization params for inputs and outputs \
                       of nodes to be quantized are required.".format(value_name))
            dqlinear_name = value_name + "_DequantizeLinear"
            dqlinear_inputs = [value_name + '_quantized',
                               scale_name,
                               zp_name]
            dequantize_node = onnx.helper.make_node("DequantizeLinear",
                                                    dqlinear_inputs,
                                                    [value_name],
                                                    dqlinear_name)
            if dequantize_node not in self.new_nodes:
                self.new_nodes.append(dequantize_node)

    def _get_dynamic_input_quantization_params(self, input_name, qType):
        """
        Create nodes for dynamic quantization of input
            parameter input_name: Name of the input.
            parameter qType: type to quantize to.
            return: scale_name, zero_point_name, scale_shape, zero_point_shape.
        """
        if qType == onnx_proto.TensorProto.INT8:
            return self._get_dynamic_input_quantization_params_int8(input_name)

        return self._get_dynamic_input_quantization_params_uint8(input_name)

    def _get_dynamic_input_quantization_params_int8(self, input_name): # pragma: no cover
        """
        Create nodes for dynamic quantization of input to int8
            parameter input_name: Name of the input.
            return: scale_name, zero_point_name, scale_shape, zero_point_shape.
        """
        qType = onnx_proto.TensorProto.INT8

        # Reduce min and Reduce max
        input_scale_name = input_name + "_scale"

        reduce_min_name = input_name + "_ReduceMin"
        reduce_min_node = onnx.helper.make_node(
            "ReduceMin",
            [input_name],
            [reduce_min_name + ":0"],
            reduce_min_name,
            keepdims=0,
        )
        self.new_nodes.append(reduce_min_node)

        reduce_max_name = input_name + "_ReduceMax"
        reduce_max_node = onnx.helper.make_node(
            "ReduceMax",
            [input_name],
            [reduce_max_name + ":0"],
            reduce_max_name,
            keepdims=0,
        )
        self.new_nodes.append(reduce_max_node)

        # Compute scale
        #   Find abs(rmin)
        reduce_min_abs_name = reduce_min_name + "_Abs"
        reduce_min_abs_node = onnx.helper.make_node(
            "Abs",
            [reduce_min_node.output[0]],
            [reduce_min_abs_name + ":0"],
            reduce_min_abs_name,
        )
        self.new_nodes.append(reduce_min_abs_node)
        #   Find abs(rmax)
        reduce_max_abs_name = reduce_max_name + "_Abs"
        reduce_max_abs_node = onnx.helper.make_node(
            "Abs",
            [reduce_max_node.output[0]],
            [reduce_max_abs_name + ":0"],
            reduce_max_abs_name,
        )
        self.new_nodes.append(reduce_max_abs_node)
        #   Compute max of abs(rmin) and abs(rmax)
        abs_max_name = input_name + "_Abs_Max"
        abs_max_node = onnx.helper.make_node(
            "Max",
            [reduce_min_abs_node.output[0], reduce_max_abs_node.output[0]],
            [abs_max_name + ":0"],
            abs_max_name,
        )
        self.new_nodes.append(abs_max_node)
        #   and divide by (quantize_range/2.0) which will be equal to max(...)*2.0/quantize_range
        initializer_div = onnx.helper.make_tensor(
            self.fixed_qrange_int8_name,
            onnx_proto.TensorProto.FLOAT,
            [],
            [_get_qrange_for_qType(qType) / 2.0],
        )
        self.model.add_initializer(initializer_div)
        scale_div_name = input_name + "scale_Div"
        scale_div_node = onnx.helper.make_node(
            "Div",
            [abs_max_node.output[0], self.fixed_qrange_int8_name],
            [input_scale_name],
            scale_div_name,
        )
        self.new_nodes.append(scale_div_node)

        # Zero point
        initializer_zp = onnx.helper.make_tensor(self.fixed_zero_zp_name, qType, [], [0])
        self.model.add_initializer(initializer_zp)

        return input_scale_name, self.fixed_zero_zp_name, [], []

    def _get_dynamic_input_quantization_params_uint8(self, input_name):
        """
        Create nodes for dynamic quantization of input to uint8
            parameter input_name: Name of the input.
            return: scale_name, zero_point_name, scale_shape, zero_point_shape.
        """
        qType = onnx_proto.TensorProto.UINT8
        # Reduce min and Reduce max
        input_scale_name = input_name + "_scale"
        input_zp_name = input_name + "_zero_point"

        reduce_min_name = input_name + "_ReduceMin"
        reduce_min_node = onnx.helper.make_node(
            "ReduceMin",
            [input_name],
            [reduce_min_name + ":0"],
            reduce_min_name,
            keepdims=0,
        )
        self.new_nodes.append(reduce_min_node)

        reduce_max_name = input_name + "_ReduceMax"
        reduce_max_node = onnx.helper.make_node(
            "ReduceMax",
            [input_name],
            [reduce_max_name + ":0"],
            reduce_max_name,
            keepdims=0,
        )
        self.new_nodes.append(reduce_max_node)

        # Add tensors for quantize range and zero value.
        initializer_qrange = onnx.helper.make_tensor(
            self.fixed_qrange_uint8_name,
            onnx_proto.TensorProto.FLOAT,
            [],
            [_get_qrange_for_qType(qType)],
        )
        self.model.add_initializer(initializer_qrange)
        initializer_qvalue = onnx.helper.make_tensor(
            self.fixed_zero_name, onnx_proto.TensorProto.FLOAT, [], [0.0])
        self.model.add_initializer(initializer_qvalue)

        # Compute Scale
        #   Subtract rmax and rmin
        scale_sub_name = input_name + "_scale_Sub"
        scale_sub_node = onnx.helper.make_node(
            "Sub",
            [reduce_max_node.output[0], reduce_min_node.output[0]],
            [scale_sub_name + ":0"],
            scale_sub_name,
        )
        self.new_nodes.append(scale_sub_node)
        #   and divide by quantize range
        scale_div_name = input_name + "_scale_Div"
        scale_div_node = onnx.helper.make_node(
            "Div",
            [scale_sub_node.output[0], self.fixed_qrange_uint8_name],
            [input_scale_name],
            scale_div_name,
        )
        self.new_nodes.append(scale_div_node)

        # Compute zero point
        #   Subtract zero and rmin
        zp_sub_name = input_name + "_zero_point_Sub"
        zp_sub_node = onnx.helper.make_node(
            "Sub",
            [self.fixed_zero_name, reduce_min_node.output[0]],
            [zp_sub_name + ":0"],
            zp_sub_name,
        )
        self.new_nodes.append(zp_sub_node)
        #   Divide by scale
        zp_div_name = input_name + "_zero_point_Div"
        zp_div_node = onnx.helper.make_node(
            "Div",
            [zp_sub_node.output[0], input_scale_name],
            [zp_div_name + ":0"],
            zp_div_name,
        )
        self.new_nodes.append(zp_div_node)
        #   Compute floor
        zp_floor_name = input_name + "_zero_point_Floor"
        zp_floor_node = onnx.helper.make_node(
            "Floor", zp_div_node.output, [zp_floor_name + ":0"], zp_floor_name)
        self.new_nodes.append(zp_floor_node)
        #   Cast to integer
        zp_cast_name = input_name + "_zero_point_Cast"
        zp_cast_node = onnx.helper.make_node(
            "Cast", zp_floor_node.output, [input_zp_name], zp_cast_name, to=qType)
        self.new_nodes.append(zp_cast_node)

        return input_scale_name, input_zp_name, [], []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Official evaluation script for v1.1 of the SQuAD dataset.
https://github.com/allenai/bi-att-flow/blob/master/squad/evaluate-v1.1.py """

import re
import six
import sys
import numpy as np
import pandas as pd
import unicodedata
from .bleu_util import compute_bleu
from .metric import metric_registry

class UnicodeRegex(object):
    def __init__(self):
        punctuation = self.property_chars("P")
        self.nondigit_punct_re = re.compile(r"([^\d])([" + punctuation + r"])")
        self.punct_nondigit_re = re.compile(r"([" + punctuation + r"])([^\d])")
        self.symbol_re = re.compile("([" + self.property_chars("S") + "])")

    def property_chars(self, prefix):
        return "".join(six.unichr(x) for x in range(sys.maxunicode)
                    if unicodedata.category(six.unichr(x)).startswith(prefix))

uregex = UnicodeRegex()

def bleu_tokenize(string):
    string = uregex.nondigit_punct_re.sub(r"\1 \2 ", string)
    string = uregex.punct_nondigit_re.sub(r" \1 \2", string)
    string = uregex.symbol_re.sub(r" \1 ", string)
    return string.split()

@metric_registry('BLEU', 'tensorflow, inteltensorflow')
class BLEU(object):
    """Computes Bilingual Evaluation Understudy Score

    BLEU score computation between labels and predictions. An approximate BLEU scoring 
    method since we do not glue word pieces or decode the ids and tokenize the output. 
    By default, we use ngram order of 4 and use brevity penalty. Also, this does not 
    have beam search
    
    """
    def __init__(self):
        self.translations = []
        self.labels = []

    def reset(self):
        """clear preds and labels storage"""
        self.translations = []
        self.labels = []

    def update(self, pred, label):
        """add preds and labels to storage"""
        if len(label) != len(pred):
            raise ValueError("Reference and translation files have different number "
                             "of lines. If training only a few steps (100-200), the "
                             "translation may be empty.")
        label = [x.lower() for x in label]
        pred = [x.lower() for x in pred]
        label = [bleu_tokenize(x) for x in label]
        pred = [bleu_tokenize(x) for x in pred]
        self.labels.extend(label)
        self.translations.extend(pred)

    def result(self):
        """calculate metric"""
        return compute_bleu(self.labels, self.translations) * 100




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .component import Component
from .quantization import Quantization
from .pruning import Pruning
from .benchmark import Benchmark
from .graph_optimization import Graph_Optimization, GraphOptimization
from .mixed_precision import MixedPrecision
from .model_conversion import ModelConversion
from .distillation import Distillation
from .nas import NAS

__all__ = ['Component', 'Quantization', 'Pruning', 'Benchmark', 'Graph_Optimization', \
           'GraphOptimization', 'ModelConversion', 'Distillation', 'NAS', 'MixedPrecision']



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pandas as pd

from neural_compressor.conf.config import Conf, NASConfig
from neural_compressor.utils import logger

from .nas import NASBase
from .nas_utils import nas_registry


@nas_registry("DyNAS")
class DyNAS(NASBase):
    """
    Args:
        conf_fname_or_obj (string or obj):
            The path to the YAML configuration file or the object of NASConfig.
    """
    def __init__(self, conf_fname_or_obj):
        from .dynast.dynas_manager import ParameterManager
        from .dynast.dynas_predictor import Predictor
        from .dynast.dynas_search import ProblemMultiObjective, SearchAlgoManager
        from .dynast.dynas_utils import (EvaluationInterfaceMobileNetV3,
                                        EvaluationInterfaceResNet50, OFARunner)
        self.ParameterManager = ParameterManager
        self.Predictor = Predictor
        self.ProblemMultiObjective = ProblemMultiObjective
        self.SearchAlgoManager = SearchAlgoManager
        self.OFARunner = OFARunner
        self.SUPERNET_PARAMETERS = {
                                    'ofa_resnet50':
                                        {'d'  :  {'count' : 5,  'vars' : [0, 1, 2]},
                                            'e'  :  {'count' : 18, 'vars' : [0.2, 0.25, 0.35]},
                                            'w'  :  {'count' : 6,  'vars' : [0, 1, 2]} },
                                    'ofa_mbv3_d234_e346_k357_w1.0':
                                        {'ks'  :  {'count' : 20, 'vars' : [3, 5, 7]},
                                            'e'   :  {'count' : 20, 'vars' : [3, 4, 6]},
                                            'd'   :  {'count' : 5,  'vars' : [2, 3, 4]} },
                                        'ofa_mbv3_d234_e346_k357_w1.2':
                                        {'ks'  :  {'count' : 20, 'vars' : [3, 5, 7]},
                                            'e'   :  {'count' : 20, 'vars' : [3, 4, 6]},
                                            'd'   :  {'count' : 5,  'vars' : [2, 3, 4]} }
                                    }
        self.EVALUATION_INTERFACE = {'ofa_resnet50': EvaluationInterfaceResNet50,
                                     'ofa_mbv3_d234_e346_k357_w1.0': EvaluationInterfaceMobileNetV3,
                                     'ofa_mbv3_d234_e346_k357_w1.2': EvaluationInterfaceMobileNetV3}
        self.LINAS_INNERLOOP_EVALS = {'ofa_resnet50': 5000,
                                      'ofa_mbv3_d234_e346_k357_w1.0': 20000,
                                      'ofa_mbv3_d234_e346_k357_w1.2': 20000}
        super().__init__()
        self.acc_predictor = None
        self.macs_predictor = None
        self.latency_predictor = None
        self.results_csv_path = None
        self.init_cfg(conf_fname_or_obj)


    def estimate(self, individual):
        self.validation_interface.eval_subnet(individual)

    def init_for_search(self):
        self.supernet_manager = self.ParameterManager(
            param_dict=self.SUPERNET_PARAMETERS[self.supernet],
            seed=self.seed
        )

        # Validation High-Fidelity Measurement Runner
        self.runner_validate = self.OFARunner(
            supernet=self.supernet,
            acc_predictor=None,
            macs_predictor=None,
            latency_predictor=None,
            imagenetpath=self.dataset_path,
            batch_size=self.batch_size,
        )

        # Setup validation interface
        self.validation_interface = self.EVALUATION_INTERFACE[self.supernet](
            evaluator=self.runner_validate,
            metrics=self.metrics,
            manager=self.supernet_manager,
            csv_path=self.results_csv_path
        )

        # Clear csv file if one exists
        # self.validation_interface.clear_csv()

    def search(self):
        self.init_for_search()

        # Randomly sample search space for initial population
        # if number of results in results_csv_path smaller than population.
        df = pd.read_csv(self.results_csv_path)
        latest_population = [self.supernet_manager.random_sample() \
            for _ in range(max(self.population - df.shape[0], 0))]

        # Start Lightweight Iterative Neural Architecture Search (LINAS)
        num_loops = round(self.num_evals/self.population)
        for loop in range(num_loops):
            logger.info('[DyNAS-T] Starting LINAS loop {} of {}.'.format(loop+1, num_loops))

            for individual in latest_population:
                self.validation_interface.eval_subnet(individual)

            self.create_acc_predictor()
            self.create_macs_predictor()
            self.create_latency_predictor()

            # Inner-loop Low-Fidelity Predictor Runner, need to re-instantiate every loop
            runner_predict = self.OFARunner(
                supernet=self.supernet,
                acc_predictor=self.acc_predictor,
                macs_predictor=self.macs_predictor,
                latency_predictor=self.latency_predictor,
                imagenetpath=self.dataset_path,
                batch_size=self.batch_size,
            )

            # Setup validation interface
            prediction_interface = self.EVALUATION_INTERFACE[self.supernet](
                evaluator=runner_predict,
                manager=self.supernet_manager,
                metrics=self.metrics,
                csv_path=None,
                predictor_mode = True
            )

            problem = self.ProblemMultiObjective(
                evaluation_interface=prediction_interface,
                param_count=self.supernet_manager.param_count,
                param_upperbound=self.supernet_manager.param_upperbound
            )

            if self.search_algo == 'age':
                search_manager = self.SearchAlgoManager(algorithm='age', seed=self.seed)
                search_manager.configure_age(population=self.population,
                                            num_evals=self.LINAS_INNERLOOP_EVALS[self.supernet])
            else:
                search_manager = self.SearchAlgoManager(algorithm='nsga2', seed=self.seed)
                search_manager.configure_nsga2(population=self.population,
                                            num_evals=self.LINAS_INNERLOOP_EVALS[self.supernet])

            results = search_manager.run_search(problem)

            latest_population = results.pop.get('X')

        logger.info("[DyNAS-T] Validated model architectures in file: {}".format(self.results_csv_path))

        output = list()
        for individual in latest_population:
            output.append(self.supernet_manager.translate2param(individual))

        return output

    def select_model_arch(self): # pragma: no cover
        # model_arch_proposition intrinsically contained in
        # pymoo.minimize API of search_manager.run_search method,
        # don't have to implement it explicitly.
        pass

    def create_acc_predictor(self):
        if 'acc' in self.metrics:
            logger.info('Building Accuracy Predictor')
            df = self.supernet_manager.import_csv(self.results_csv_path,
                                                  config='config',
                                                  objective='acc',
                                                  column_names=['config','date','lat','macs','acc'])
            features, labels = self.supernet_manager.create_training_set(df)
            self.acc_predictor = self.Predictor()
            self.acc_predictor.train(features, labels.ravel())
        else:
            self.acc_predictor = None

    def create_macs_predictor(self):
        if 'macs' in self.metrics:
            logger.info('Building MACs Predictor')
            df = self.supernet_manager.import_csv(self.results_csv_path,
                                                  config='config',
                                                  objective='macs',
                                                  column_names=['config','date','lat','macs','acc'])
            features, labels = self.supernet_manager.create_training_set(df)
            self.macs_predictor = self.Predictor()
            self.macs_predictor.train(features, labels.ravel())
        else:
            self.macs_predictor = None

    def create_latency_predictor(self):
        if 'lat' in self.metrics:
            logger.info('Building Latency Predictor')
            df = self.supernet_manager.import_csv(self.results_csv_path,
                                                  config='config',
                                                  objective='lat',
                                                  column_names=['config','date','lat','macs','acc'])
            features, labels = self.supernet_manager.create_training_set(df)
            self.latency_predictor = self.Predictor()
            self.latency_predictor.train(features, labels.ravel())
        else:
            self.latency_predictor = None

    def init_cfg(self, conf_fname_or_obj):
        if isinstance(conf_fname_or_obj, str):
            if os.path.isfile(conf_fname_or_obj):
                self.conf = Conf(conf_fname_or_obj).usr_cfg
        elif isinstance(conf_fname_or_obj, NASConfig):
            conf_fname_or_obj.validate()
            self.conf = conf_fname_or_obj.usr_cfg
        else: # pragma: no cover
            raise NotImplementedError(
                "Please provide a str path to the config file or an object of NASConfig."
            )
        #self.init_search_cfg(self.conf.nas)
        assert 'dynas' in self.conf.nas, "Must specify dynas section."
        dynas_config = self.conf.nas.dynas
        self.search_algo = self.conf.nas.search.search_algorithm
        self.supernet = dynas_config.supernet
        self.metrics = dynas_config.metrics
        self.num_evals = dynas_config.num_evals
        self.results_csv_path = dynas_config.results_csv_path
        self.dataset_path = dynas_config.dataset_path
        self.batch_size = dynas_config.batch_size
        if dynas_config.population < 10: # pragma: no cover
            raise NotImplementedError(
                "Please specify a population size >= 10"
            )
        else:
            self.population = dynas_config.population



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .dataset import dataset_registry, Dataset
import numpy as np
from neural_compressor.utils.utility import LazyImport
import logging

mx = LazyImport('mxnet')
torch = LazyImport('torch')

logger = logging.getLogger()

@dataset_registry(dataset_type="dummy", framework="tensorflow, inteltensorflow, tensorflow_itex, \
                                                   onnxrt_qlinearops, onnxrt_integerops, \
                                                   pytorch, pytorch_ipex, pytorch_fx, \
                                                   mxnet",
                                                   dataset_format='')
class DummyDataset(Dataset):
    """Dataset used for dummy data generation.
       This Dataset is to construct a dataset from a specific shape.
       the value range is calculated from: low * stand_normal(0, 1) + high
       (TODO) construct dummy data from real dataset or iteration of data.

    Args: shape (list or tuple):support create multi shape tensors, use list of tuples
                                for each tuple in the list, will create a such size tensor.
          low (list or float, default=-128.):low out the tensor value range from [0, 1] 
                                            to [0, low] or [low, 0] if low < 0, if float, 
                                            will implement all tensors with same low value.  
          high (list or float, default=127.):high the tensor value by add all tensor element
                                            value high. If list, length of list should be 
                                            same with shape list.
          dtype (list or str, default='float32'):support multi tensor dtype setting. If list,
                                                length of list should be same with shape list,
                                                if str, all tensors will use same dtype. dtype
                                                support 'float32', 'float16', 'uint8', 'int8',
                                                'int32', 'int64', 'bool'.
          label (bool, default=True):whether to return 0 as label.
          transform (transform object, default=None): dummy dataset does not need transform.
                                                        If transform is not None, it will ignore
                                                        it.  
          filter (Filter objects, default=None): filter out examples according to 
                                                specific conditions                                         
    """
    def __init__(self, shape, low=-128., high=127., dtype='float32', label=True, \
        transform=None, filter=None):

        dtype_map = {'float32':np.float32, 'float16':np.float16, 'uint8':np.uint8, \
                     'int8':np.int8, 'int32':np.int32, 'int64':np.int64, 'bool':np.bool}

        np.random.seed(9527)
        self.transform = transform
        self.label = label
        if len(shape)==0:
            logger.info("No data in the dummy dataset.")
        elif isinstance(shape, list):
            # list tensor should same first demension n
            n = shape[0][0]
            assert all(isinstance(elem, tuple) and elem[0] == n for elem in shape), \
                'each tensor shape should be tuple and same fisrt demension'

            if isinstance(low, list):
                assert len(low) == len(shape) and all(isinstance(elem, float) for elem in low), \
                    'low list should have same length with shape with element data type float'
            else:
                low = (low * np.ones(len(shape))).astype(np.float)

            if isinstance(high, list):
                assert len(high) == len(shape) and all(isinstance(elem, float) for elem in high), \
                    'high list should have same length with shape with element data type float'
            else:
                high = (high * np.ones(len(shape))).astype(np.float)

            if isinstance(dtype, list):
                assert len(dtype) == len(shape) and \
                    all(elem in dtype_map.keys() for elem in dtype), \
                    'high list should have same length with shape with element data type float'
            else:
                dtype = [dtype for i in range(0, len(shape))]

        elif isinstance(shape, tuple):
            shape = [shape]
            if isinstance(low, float):
                low = [low]
            else:
                assert isinstance(low, list) and len(low) == 1 and isinstance(low[0], float), \
                    'low should be float or list of float with length 1'

            if isinstance(high, float):
                high = [high]
            else:
                assert isinstance(high, list) and len(high) == 1 and isinstance(high[0], float), \
                    'high should be float or list of float with length 1'

            if isinstance(dtype, str):
                assert dtype in dtype_map.keys(), 'dtype only support {}'.format(dtype_map.keys())
                dtype = [dtype]
            else:
                assert isinstance(dtype, list) and \
                    len(dtype) == 1 and dtype[0] in dtype_map.keys(), \
                    'dtype should be str or list of str in supported dtypes'

        self.dataset = []
        for idx in range(0, len(shape)):
            tensor = np.random.uniform(low=low[idx], high=high[idx], size=shape[idx])
            tensor = tensor.astype(dtype_map[dtype[idx]])
            self.dataset.append(tensor)

        if len(self.dataset) == 1:
            self.dataset = self.dataset[0]
        else:
            self.dataset = [elem for elem in zip(*self.dataset)]


    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):

        sample = self.dataset[index]
        if self.transform is not None:
            logger.warning("Dummy dataset does not need transform.")

        if self.label:
            return sample, 0
        else:
            return sample



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from tensorflow.python.framework import dtypes
from tensorflow.core.framework import node_def_pb2
from ..quantize_graph_base import QuantizeNodeBase
from neural_compressor.adaptor.tf_utils.quantize_graph_common import QuantizeGraphHelper as helper


class FuseNodeStartWithConcatV2(QuantizeNodeBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.sorted_patterns = sorted(self.patterns,
                                      key=lambda i: len(i),
                                      reverse=True)

    def _quantizable_concat(self, node):
        deq_type = []
        for input_node_name in node.input[:node.attr['N'].i]:
            node_name = helper.node_name_from_input(input_node_name)
            if self.node_name_mapping[node_name].node.op != "Dequantize":
                return False

            deq_type.append(self.node_name_mapping[node_name].node.attr['T'].type)

        if len(set(deq_type)) != 1:
            return False

        return True

    def _apply_concatv2_quantization(self, match_node_name):
        # Dequantize + ConcatV2 + QuantizeV2
        skip_node_names = match_node_name[2:]
        matched_node = self.node_name_mapping[match_node_name[1]]
        _, normal_inputs = self._get_node_input(matched_node.node.name)
        num_input = len(normal_inputs)
        shape_input_name = normal_inputs[num_input - 1]
        original_inputs = normal_inputs[0:num_input - 1]

        input_names = []
        min_names = []
        max_names = []
        for each_input in original_inputs:
            _, q_each_input = self._get_node_input(each_input)
            input_names.append(q_each_input[0])
            min_names.append(q_each_input[1])
            max_names.append(q_each_input[2])
            skip_node_names.append(each_input)

        all_input_names = input_names
        all_input_names.append(shape_input_name)
        all_input_names.extend(min_names)
        all_input_names.extend(max_names)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_names:
                self.logger.debug("skip node {}".format(node.name))
            elif node.name == match_node_name[1]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))
                quantized_op_name = node.name + "_eightbit_quantized_concatv2"
                quantized_op_type = "Quantized" + node.op
                quantized_concat_node = helper.create_node(quantized_op_type, quantized_op_name, all_input_names)

                helper.set_attr_int(quantized_concat_node, "N", len(original_inputs))
                helper.set_attr_dtype(quantized_concat_node, "T", dtypes.quint8)
                self.add_output_graph_node(quantized_concat_node)
                self._intel_cpu_add_dequantize_result_node(quantized_op_name, node.name)
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def get_longest_fuse(self):
        self._get_op_list()
        matched_node_name = []
        
        for k, v in enumerate(self.op_list):
            if v in set(fusion[1] for fusion in self.sorted_patterns):
                cur_node = self.node_name_mapping[list(
                    self.node_name_mapping.keys())[k]].node

                if cur_node.name != self.start_node_name:
                    continue

                for sub_rule in self.sorted_patterns:
                    if sub_rule[0] != "Dequantize" or sub_rule[-1] != "QuantizeV2":
                        continue
                    if v != sub_rule[1]:
                        continue

                if self._quantizable_concat(cur_node) and \
                    dtypes.as_dtype(cur_node.attr["T"].type) == dtypes.float32 and \
                    not re.search(r'map(_\d+)?/while', cur_node.name):
                    matched_node_name.clear()
                    matched_node_name.append(sub_rule[0])
                    matched_node_name.append(cur_node.name)
                    matched_node_name.append(sub_rule[-1])
                    return sub_rule, matched_node_name

        return None, None

    def apply_the_transform(self):
        self._get_op_list()
        matched_rule, matched_node_name = self.get_longest_fuse()
        if matched_node_name:
            fusion_name = ''.join(matched_rule)
            if fusion_name == "DequantizeConcatV2QuantizeV2":
                self._apply_concatv2_quantization(matched_node_name)
            else: # pragma: no cover
                self.logger.info("Unknown fusion pattern {}.".format(fusion_name))
                if self.remove_redundant_quant_flag:
                    self.input_graph = self.remove_redundant_quantization(self.input_graph)
                return self.input_graph

            self.input_graph = self.output_graph
            self._reset_output_node_maps()
            if self.remove_redundant_quant_flag:
                self.output_graph = self.remove_redundant_quantization(self.output_graph)
            return self.output_graph

        if self.remove_redundant_quant_flag:
            self.input_graph = self.remove_redundant_quantization(self.input_graph)
        return self.input_graph



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf
from tensorflow.core.framework import graph_pb2
from tensorflow.core.framework import node_def_pb2
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import tensor_util

from neural_compressor.adaptor.tf_utils.quantize_graph_common import QuantizeGraphHelper as helper
from .quantize_graph_base import QuantizeNodeBase
from neural_compressor.adaptor.tf_utils.util import version1_gte_version2
import numpy as np

class FuseNodeStartWithConv2d(QuantizeNodeBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.sorted_patterns = sorted(self.patterns,
                                      key=lambda i: len(i),
                                      reverse=True)

        self.fusion_mapping = {
            'Conv2DBiasAdd': self.apply_conv_biasadd_fusion,
            'Conv2DBiasAddAddNRelu': self.apply_conv_biasadd_addn_relu_fusion,
            'Conv2DBiasAddAddNRelu6': self.apply_conv_biasadd_addn_relu_fusion,
            'Conv2DBiasAddAddV2Relu': self.apply_conv_biasadd_addn_relu_fusion,
            'Conv2DBiasAddAddV2Relu6': self.apply_conv_biasadd_addn_relu_fusion,
            'Conv2DBiasAddAddRelu': self.apply_conv_biasadd_addn_relu_fusion,
            'Conv2DBiasAddRelu6': self.apply_conv_biasadd_relu_fusion,
            'Conv2DBiasAddRelu': self.apply_conv_biasadd_relu_fusion,
            'Conv2DBiasAddLeakyRelu': self.apply_conv_biasadd_relu_fusion,
            'Conv2DBiasAddLeakyReluAddV2': self.apply_conv_biasadd_addn_relu_fusion,
            'Conv2DAddRelu6': self.apply_conv_biasadd_relu_fusion,
            'Conv2DAddRelu': self.apply_conv_biasadd_relu_fusion,
            'DepthwiseConv2dNativeAddRelu6':
            self.apply_conv_biasadd_relu_fusion,
            'DepthwiseConv2dNativeBiasAddRelu':
            self.apply_conv_biasadd_relu_fusion,
            'DepthwiseConv2dNativeBiasAdd': self.apply_conv_biasadd_fusion,
            'DepthwiseConv2dNativeBiasAddRelu6':
            self.apply_conv_biasadd_relu_fusion,
            'Conv2D': self.apply_conv_single_fusion,
            'DepthwiseConv2dNative': self.apply_conv_single_fusion
        }

    def apply_conv_single_fusion(self, match_node_name):
        skip_node_name = match_node_name[1:]
        matched_node = self.node_name_mapping[match_node_name[0]]
        _, normal_inputs = self._get_node_input(matched_node.node.name)
        weight_name = normal_inputs[1]
        # TODO this is workaround as the tf 2.1 doesn't support depthwise/conv s8
        # feature.
        if self.enable_s8 and not self._find_relu_node(matched_node.node):
            self.output_graph = self.input_graph
            return

        q_weights_name, q_weights_min_name, q_weights_max_name = \
            self._intel_cpu_quantize_weight_eightbit(matched_node.node.op,
                                                    self.node_name_mapping[weight_name].node,
                                                    self.per_channel)

        all_input_names = self._add_eightbit_prologue_nodes(matched_node.node.name)
        all_input_names = all_input_names[:1] + [q_weights_name] + all_input_names[1:]
        all_input_names.append(q_weights_min_name)
        all_input_names.append(q_weights_max_name)
        skip_node_name.append(weight_name)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                self.logger.debug("Skip node {}.".format(node.name))
            elif node.name == match_node_name[0]:
                postfix = "_eightbit_quantized_depthwise_conv"
                if node.op == "Conv2D":
                    postfix = "_eightbit_quantized_conv"
                quantized_node_name = node.name + postfix
                if node.op == "Conv2D":
                    quantized_conv_node = helper.create_node(
                        "QuantizedConv2DPerChannel"
                        if self.per_channel else "QuantizedConv2D",
                        quantized_node_name, all_input_names)

                elif node.op == "DepthwiseConv2dNative":
                    quantized_conv_node = helper.create_node(
                        "QuantizedDepthwiseConv2D", quantized_node_name,
                        all_input_names)

                helper.copy_attr(quantized_conv_node, "strides",
                                 node.attr["strides"])
                helper.copy_attr(quantized_conv_node, "padding",
                                 node.attr["padding"])
                if "padding_list" in node.attr:
                    helper.copy_attr(quantized_conv_node, "padding_list",
                                     node.attr["padding_list"])
                helper.copy_attr(quantized_conv_node, "dilations",
                                 node.attr["dilations"])
                input_data_type = dtypes.quint8 if self._find_relu_node(
                    node) else dtypes.qint8
                helper.set_attr_dtype(quantized_conv_node, "Tinput",
                                      input_data_type)
                helper.set_attr_dtype(quantized_conv_node, "Tfilter",
                                      dtypes.qint8)
                helper.set_attr_dtype(quantized_conv_node, "out_type",
                                      dtypes.qint32)
                self.add_output_graph_node(quantized_conv_node)
                quantize_down_name = self._add_quantize_down_nodes(
                    node, quantized_node_name, dtypes.qint8)
                self._intel_cpu_add_dequantize_result_node(
                    quantize_down_name, node.name, dtypes.qint8)
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def apply_conv_biasadd_relu_fusion(self, match_node_name):
        """Fuse the conv/biasadd/relu pattern.

        Arguments:
            match_node_name {[type]} -- [description]
        """

        skip_node_name = match_node_name[1:]
        matched_node = self.node_name_mapping[match_node_name[0]]
        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)
        weight_name = normal_inputs[1]

        q_weights_name, q_weights_min_name, q_weights_max_name = \
            self._intel_cpu_quantize_weight_eightbit(
                matched_node.node.op, self.node_name_mapping[weight_name].node, self.per_channel)

        all_input_names = self._add_eightbit_prologue_nodes(matched_node.node.name)
        all_input_names = all_input_names[:1] + [q_weights_name] + all_input_names[1:]
        all_input_names.append(q_weights_min_name)
        all_input_names.append(q_weights_max_name)
        skip_node_name.append(weight_name)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                self.logger.debug("Skip node {}.".format(node.name))
            elif node.name == match_node_name[0]:

                postfix = "_eightbit_quantized_depthwise_conv"
                if node.op == "Conv2D":
                    postfix = "_eightbit_quantized_conv"
                quantized_node_name = node.name + postfix
                bias_node_name = self.node_name_mapping[match_node_name[1]].node.input[1]
                relu_node_name = match_node_name[2]
                is_relu6 = self.node_name_mapping[relu_node_name].node.op == "Relu6"
                quantized_node_input_names = all_input_names[:2] + \
                    [bias_node_name] + all_input_names[2:] + control_inputs
                is_leakyrelu = self.node_name_mapping[relu_node_name].node.op == "LeakyRelu"
                quantized_conv_node_op = 'QuantizedDepthwiseConv2DWithBiasAndRelu'
                if node.op == "Conv2D" or is_leakyrelu:
                    quantized_conv_node_op = "QuantizedConv2DWithBiasAndRelu"
                quantized_conv_node = helper.create_node(
                    quantized_conv_node_op,
                    quantized_node_name, quantized_node_input_names)
                helper.copy_attr(quantized_conv_node, "strides", node.attr["strides"])
                helper.copy_attr(quantized_conv_node, "padding", node.attr["padding"])
                if "alpha" in self.node_name_mapping[relu_node_name].node.attr:
                    helper.copy_attr(quantized_conv_node, "alpha",
                    self.node_name_mapping[relu_node_name].node.attr["alpha"])
                if "padding_list" in node.attr:
                    helper.copy_attr(quantized_conv_node, "padding_list",
                    node.attr["padding_list"])
                helper.copy_attr(quantized_conv_node, "dilations", node.attr["dilations"])
                input_data_type = dtypes.quint8 if self._find_relu_node(node) else dtypes.qint8
                helper.set_attr_dtype(quantized_conv_node, "Tinput", input_data_type)
                helper.set_attr_dtype(quantized_conv_node, "Tfilter",dtypes.qint8)
                helper.set_attr_dtype(quantized_conv_node, "out_type", dtypes.qint32)
                self.add_output_graph_node(quantized_conv_node)
                if not is_leakyrelu:
                    quantize_down_name = self._add_quantize_down_nodes(
                        node, quantized_node_name, dtypes.quint8, is_relu6)
                    self._intel_cpu_add_dequantize_result_node(
                        quantize_down_name, relu_node_name)
                else:
                    quantize_down_name = self._add_quantize_down_nodes(
                        node, quantized_node_name, dtypes.qint8, False)
                    self._intel_cpu_add_dequantize_result_node(
                        quantize_down_name, relu_node_name, dtype=dtypes.qint8)
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def apply_conv_biasadd_fusion(self, match_node_name):
        skip_node_name = match_node_name[1:]
        matched_node = self.node_name_mapping[match_node_name[0]]
        control_inputs, normal_inputs = self._get_node_input(
            matched_node.node.name)
        weight_name = normal_inputs[1]

        q_weights_name, q_weights_min_name, q_weights_max_name = \
            self._intel_cpu_quantize_weight_eightbit(
                matched_node.node.op, self.node_name_mapping[weight_name].node, self.per_channel)

        all_input_names = self._add_eightbit_prologue_nodes(matched_node.node.name)
        all_input_names = all_input_names[:1] + [q_weights_name] + all_input_names[1:]
        all_input_names.append(q_weights_min_name)
        all_input_names.append(q_weights_max_name)
        skip_node_name.append(weight_name)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                pass
            elif node.name == match_node_name[0]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))

                quantized_node_name = node.name + "_eightbit_quantized_conv"
                if node.op == "DepthwiseConv2dNative":
                    quantized_node_name = node.name + "_eightbit_quantized_depthwise_conv"

                bias_node_name = self.node_name_mapping[
                    match_node_name[1]].node.input[1]
                quantized_node_input_names = all_input_names[:2] + [
                    bias_node_name
                ] + all_input_names[2:] + control_inputs

                quantized_conv_node = helper.create_node(
                    "QuantizedConv2DWithBias" if node.op == 'Conv2D' \
                        else 'QuantizedDepthwiseConv2DWithBias',
                    quantized_node_name,
                    quantized_node_input_names)

                helper.copy_attr(quantized_conv_node, "strides", node.attr["strides"])
                helper.copy_attr(quantized_conv_node, "padding", node.attr["padding"])
                if "padding_list" in node.attr:
                    helper.copy_attr(quantized_conv_node, "padding_list", node.attr["padding_list"])
                helper.copy_attr(quantized_conv_node, "dilations", node.attr["dilations"])

                input_data_type = dtypes.quint8 if self._find_relu_node(node) else dtypes.qint8

                helper.set_attr_dtype(quantized_conv_node, "Tinput", input_data_type)
                helper.set_attr_dtype(quantized_conv_node, "Tfilter", dtypes.qint8)
                helper.set_attr_dtype(quantized_conv_node, "out_type", dtypes.qint32)
                self.add_output_graph_node(quantized_conv_node)
                requantize_type = dtypes.qint8

                quantize_down_name = self._add_quantize_down_nodes(
                    node, quantized_node_name, requantize_type, False)
                self._intel_cpu_add_dequantize_result_node(
                    quantize_down_name, match_node_name[1], requantize_type)
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def apply_conv_biasadd_addn_relu_fusion(self, match_node_name):
        skip_node_name = match_node_name[1:]
        matched_node = self.node_name_mapping[match_node_name[0]]
        control_inputs, normal_inputs = self._get_node_input(
            matched_node.node.name)
        weight_name = normal_inputs[1]

        third_node = self.node_name_mapping[match_node_name[2]].node
        forth_node = self.node_name_mapping[match_node_name[3]].node
        if third_node.op != 'LeakyRelu' and not self._find_relu_node(matched_node.node):
            return self.apply_conv_biasadd_fusion(match_node_name[:2])

        is_leakyrelu_add_fusion = third_node.op == 'LeakyRelu' and forth_node.op.find('Add') != -1

        q_weights_name, q_weights_min_name, q_weights_max_name = \
            self._intel_cpu_quantize_weight_eightbit(
                matched_node.node.op, self.node_name_mapping[weight_name].node, self.per_channel)

        all_input_names = self._add_eightbit_prologue_nodes(matched_node.node.name)
        all_input_names = all_input_names[:1] + [q_weights_name] + all_input_names[1:]
        all_input_names.append(q_weights_min_name)
        all_input_names.append(q_weights_max_name)
        skip_node_name.append(weight_name)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                self.logger.debug("skip node {}".format(node.name))
            elif node.name == match_node_name[0]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))

                quantized_node_name = node.name + "_eightbit_quantized_conv"

                bias_node_name = self.node_name_mapping[match_node_name[1]].node.input[1]
                leaky_offset = 1 if is_leakyrelu_add_fusion else 0
                if is_leakyrelu_add_fusion:
                    relu_node_name = match_node_name[2]
                else:
                    relu_node_name = match_node_name[3]
                is_relu6 = self.node_name_mapping[relu_node_name].node.op == "Relu6"

                sum_index = 1 if match_node_name[1 + leaky_offset] == self.node_name_mapping[
                    match_node_name[2 + leaky_offset]].node.input[0] else 0
                quantized_node_input_names = all_input_names[:2] + [
                    bias_node_name
                ] + all_input_names[2:] + [
                    self.node_name_mapping[
                        match_node_name[2 + leaky_offset]].node.input[sum_index]
                ] + control_inputs

                node_op = "QuantizedConv2DWithBiasReluAndSum" if is_leakyrelu_add_fusion \
                            else "QuantizedConv2DWithBiasSumAndRelu"

                quantized_conv_node = helper.create_node(node_op, quantized_node_name,
                    quantized_node_input_names)
                helper.copy_attr(quantized_conv_node, "strides", node.attr["strides"])
                helper.copy_attr(quantized_conv_node, "padding", node.attr["padding"])
                if "padding_list" in node.attr:
                    helper.copy_attr(quantized_conv_node, "padding_list",
                                     node.attr["padding_list"])
                helper.copy_attr(quantized_conv_node, "dilations", node.attr["dilations"])
                input_data_type = dtypes.quint8 if self._find_relu_node(
                    node) else dtypes.qint8
                helper.set_attr_dtype(quantized_conv_node, "Tinput", input_data_type)
                helper.set_attr_dtype(quantized_conv_node, "Tfilter", dtypes.qint8)
                helper.set_attr_dtype(quantized_conv_node, "out_type", dtypes.qint32)
                if "alpha" in self.node_name_mapping[relu_node_name].node.attr:
                    helper.copy_attr(quantized_conv_node, "alpha",
                    self.node_name_mapping[relu_node_name].node.attr["alpha"])
                self.add_output_graph_node(quantized_conv_node)

                if is_leakyrelu_add_fusion:
                    quantize_down_name = self._add_quantize_down_nodes(
                                        node, quantized_node_name, dtypes.qint8, False)
                    self._intel_cpu_add_dequantize_result_node(
                        quantize_down_name, match_node_name[3], dtype=dtypes.qint8)
                else:
                    dtype = dtypes.quint8
                    if [i for i in self.node_name_mapping[relu_node_name].output \
                        if 'FusedBatchNorm' in self.node_name_mapping[i].node.op and \
                            i in self.op_wise_config_name_list]:
                            dtype = dtypes.qint8
                    quantize_down_name = self._add_quantize_down_nodes(
                        node, quantized_node_name, dtype, is_relu6)
                    self._intel_cpu_add_dequantize_result_node(
                        quantize_down_name, relu_node_name, dtype)
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def get_longest_fuse(self):
        self._get_op_list()

        matched_rule, matched_node_name = self._is_match(self.sorted_patterns)
        return matched_rule, matched_node_name

    def apply_the_transform(self):
        self._get_op_list()
        matched_rule, matched_node_name = self._is_match(self.sorted_patterns)
        if matched_node_name:
            self.output_graph = graph_pb2.GraphDef()
            fusion_name = ''.join(matched_rule)
            if fusion_name in self.fusion_mapping:
                if fusion_name.find('Conv2DAddRelu') != -1:
                    for input_name in self.node_name_mapping[matched_node_name[1]].node.input:
                        input_node_name = helper.node_name_from_input(input_name)
                        if input_node_name != matched_node_name[0]:
                            add_const_input_node = self.node_name_mapping[input_node_name].node
                            add_node_content = tensor_util.MakeNdarray(
                                add_const_input_node.attr["value"].tensor)
                            if add_node_content.ndim != 1:
                                fusion_name = 'Conv2D'
                                matched_node_name = matched_node_name[:1]
                self.fusion_mapping[fusion_name](matched_node_name)
            else: # pragma: no cover
                self.logger.info("Unknown fusion pattern {}.".format(fusion_name))
                if self.remove_redundant_quant_flag:
                    self.input_graph = self.remove_redundant_quantization(self.input_graph)
                return self.input_graph, []

            self.input_graph = self.output_graph
            self._reset_output_node_maps()
            if self.remove_redundant_quant_flag:
                self.output_graph = self.remove_redundant_quantization(self.output_graph)

            return self.output_graph, matched_node_name

        return self.input_graph, []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import onnx
from .base_operator import QuantOperatorBase
from onnxruntime.quantization.quant_utils import QuantizedValueType, attribute_to_kwarg
from onnx import onnx_pb as onnx_proto
from neural_compressor.adaptor.ox_utils.util import QuantizedValue
'''
    Quantize Gather
'''

class GatherConverter(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def convert(self):
        node = self.node
        assert node.op_type in ["Gather"]

        if len(self.quantizer.model.get_children(node)) == 0:
            return
        parents = self.quantizer.model.get_parents(node)
        children = self.quantizer.model.get_children(node)

        if len(parents) == 0:
            return
        if any([i.op_type == 'DequantizeLinear' for i in parents]):
            inputs = []
            inputs.append(parents[0].input[0])
            inputs.append(node.input[1])

            gather_new_output = node.output[0] + "_quantized"

            kwargs = {}
            for attribute in node.attribute:
                kwargs.update(attribute_to_kwarg(attribute))

            gather_node = onnx.helper.make_node("Gather",
                                                inputs,
                                                [gather_new_output],
                                                node.name,
                                                **kwargs)
            self.quantizer.new_nodes.append(gather_node)
            if any([i.op_type  != 'QuantizeLinear' for i in children]):
                dq_inputs = []
                dq_inputs.append(gather_new_output)
                dq_inputs.extend(parents[0].input[1:])
                dq_node = onnx.helper.make_node("DequantizeLinear",
                                                dq_inputs,
                                                [node.output[0]],
                                                node.name + '_DequantizeLinear')
                self.quantizer.new_nodes.append(dq_node)
                
            for child in children:
                if child.op_type == 'QuantizeLinear':
                    self.quantizer.remove_nodes.append(child)
                    for n in self.quantizer.model.get_children(child):
                        self.quantizer.model.replace_node_input(n, 
                                        child.output[0], gather_new_output)
            self.quantizer.remove_nodes.extend([node, parents[0]])

class GatherQuant(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def quantize(self):
        node = self.node
        assert (node.op_type == "Gather")

        if not self.quantizer.is_valid_quantize_weight(node.input[0]):
            return
        self.quantizer.quantize_inputs(node, [0])
        if not self.disable_qdq_for_node_output or self.quantizer != 'qdq':
            self.quantizer.quantize_outputs(node)
        node.name = node.name + "_quant"
